/* edit.c */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "waved.h"
#include "video.h"
#include "memory.h"

#define ESC 27

char   *marktext();
int     sigsize();
int     newmark();
int     menu();

void    markquery();
void    copyseg();
void    delseg();
void    cutseg();
void    pasteseg();
void    saveseg();
void    nsrtwave(int, int, int);
void    delewave();
void    ordmrk();
void    nsrtzero();
void    sigdel();
void    windinit();
void    blkxtnd();
void    editsamp();
void    replace();
void    update_cursor();
void    copy_rate();

void
editquery()
{
    int     c;
    static char *mis[] = {
    	"cuT", "Copy", "Delete", "Paste", "Replace", "Silence", "Mark", 
    	"Value", "eXtend", ""
    };
    
    c = menu("EDIT", mis);

    switch (c) {
    case 'B':
	blankmarks(aw);
        showdata();
	break;
    case 'C':
	copyseg();
	break;
    case 'D':
	delseg();
	break;
    case 'T':
	cutseg();
	break;
    case 'P':
	pasteseg();
	break;
    case 'R':
	replace();
	break;
    case 'S':
	nsrtzero();
	break;
    case 'X':
	blkxtnd();
	break;
    case 'V':
	editsamp();
	break;
    case 'M':
	markquery();
	break;
    }
}

/* copyseg - copy to a clip window */

void
copyseg()
{
    if (getwinmrk(" COPY:", 0)) {
	win_num = clpwin;
        if (win_num != aw) {
            ordmrk();
            sigdel(win_num);
            copy_rate(win_num, aw);
            saveseg(win_num, seg_mark1, seg_mark2);
            showwindow(win_num);
        }
    }
    clrmsg();
}

/* delseg - delete a segment */

void
delseg()
{
    if (sig[aw].siglen == 0)
        return;
    if (getwinmrk(" DELETE:", 0)) {
	ordmrk();
	if (seg_mark1 == HOME_MARK && seg_mark2 == END_MARK) {
            sigdel(aw);
	} else {
            delewave(aw, seg_mark1, seg_mark2);
        }
        showdata();
    }
    clrmsg();
}

/* cutseg - extract segment to clip window */

void
cutseg()
{
    if (sig[aw].siglen == 0)
        return;
    if (getwinmrk(" CUT:", 1)) {
	win_num = clpwin;
	if (win_num != aw) {
            ordmrk();
            sigdel(win_num);
            copy_rate(win_num, aw);
	    saveseg(win_num, seg_mark1, seg_mark2);
            delewave(aw, seg_mark1, seg_mark2);
            showdata();
            showwindow(win_num);
	}
    }
    clrmsg();
}

/* pasteseg - paste the clip segment */

void
pasteseg()
{
    int     s1b, s2b, s1z, s2z;

    if (sig[aw].siglen == 0) {
        seg_mark1 = HOME_MARK;
        seg_mark2 = END_MARK;
	sig[aw].sgbits = sig[win_num].sgbits;
	sig[aw].sigzer = sig[win_num].sigzer;
	sig[aw].sgnchn = sig[win_num].sgnchn;
    } else {
        putmsg(" PASTE: in front of mark");
        seg_mark1 = newmark();
        seg_mark2 = getmark(0);
    }
    win_num = clpwin;
    if (seg_mark2 != NOMARK && aw != win_num) {
	s1b = sig[win_num].sgbits;
	s1z = sig[win_num].sigzer;
	s2b = sig[aw].sgbits;
	s2z = sig[aw].sigzer;
	if (s2b <= 0) {
	    sig[aw].sgbits = s1b;
	    sig[aw].sigzer = s1z;
	} else if (s1b != s2b || s1z != s2z) {
	    if (sigcon(win_num, s2b, s2z, 1))
		showwindow(win_num);
	}
	nsrtwave(win_num, seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* replace - replace contents of current window with the clip signal */

void
replace()
{
    if (sig[clpwin].siglen == 0) {
        putmsgwk(" No signal in clip window. ");
    } else if (sig[aw].siglen == 0) {
        pasteseg();
    } else {
        sigdel(aw);
        copy_rate(aw, clpwin);
        pasteseg();
    }
}

/* saveseg - save segment to another window */

void
saveseg(int num, int from, int to)
{
    int     i, nchn;
    long    left, right, npts, n;
    SIGPTR  pbuf1, pbuf2;
    SIGNAL *s1, *s2;

    if (num == aw)
	return;
    s1 = &sig[aw];
    s2 = &sig[num];
    if (s1 == NULL || s2 == NULL)
	return;
    left = s1->sigmrk[from];
    right = s1->sigmrk[to];
    npts = right - left;
    nchn = limit(1, s1->sgnchn, MAX_CHAN);
    if (sigsize(num, npts, nchn) == 0)
	return;

    pbuf1 = s1->sigbuf + left * nchn;
    pbuf2 = s2->sigbuf;
    n = npts * nchn;
    while (n-- > 0)
	*pbuf2++ = *pbuf1++;

    s2->siglen = npts;
    s2->sgtime = time(NULL);
    s2->sgnchn = s1->sgnchn;
    s2->sgbits = s1->sgbits;
    s2->sigzer = s1->sigzer;
    s2->sigfmt = s1->sigfmt;
    s2->adrate = s1->adrate;
    s2->darate = s1->darate;
    for (i = 0; i < nchn; i++)
        s2->sigbuf[s2->siglen * nchn + i] = s2->sigzer;
    sprintf(s2->sigfnm, "[copy of w%d(%ld:%ld)]", aw, left, right);
    blankmarks(num);
    win[num].winmrk1 = 0;
    win[num].winmrk2 = npts;
}

void
nsrtwave(int num, int from, int to)
{
    int     c, i, nchn;
    long    n, n1, tp, n2;
    SIGPTR  pbuf1, pbuf2;
    SIGNAL *s1, *s2;

    if (num == aw)
	return;
    s1 = &sig[num];
    s2 = &sig[aw];
    if (s1 == NULL || s2 == NULL)
	return;
    if (s2->sigmrk[to] == NOMARK)
	return;
    if (s1->sgnchn != s2->sgnchn)
	return;
    nchn = limit(1, s1->sgnchn, MAX_CHAN);
    n1 = s1->siglen;
    n2 = s2->siglen;
    if (sigsize(aw, n1 + n2, nchn) == 0)
	return;
    tp = s2->sigmrk[to];
    pbuf1 = s2->sigbuf + (n2 - 1) * nchn;
    pbuf2 = s2->sigbuf + (n2 - 1 + n1) * nchn;
    n = (n2 - tp) * nchn;
    while (n-- > 0)
	*pbuf2-- = *pbuf1--;
    pbuf1 = s2->sigbuf + tp * nchn;
    pbuf2 = s1->sigbuf;
    n = n1 * nchn;
    while (n-- > 0)
	*pbuf1++ = *pbuf2++;

    s2->siglen = n1 + n2;
    for (c = 0; c < NMARK; c++)
	if (s2->sigmrk[c] != NOMARK && s2->sigmrk[c] >= tp)
	    s2->sigmrk[c] += n1;
    if (from != NOMARK)
        s2->sigmrk[from] = tp;
    if (curspnt > tp)
	curspnt += n1;
    for (i = 0; i < nchn; i++)
        s2->sigbuf[s2->siglen * nchn + i] = s2->sigzer;

    if (win[aw].winmrk1 != NOMARK && win[aw].winmrk1 == tp)
        win[aw].winmrk1 -= n1;
    if (win[aw].winmrk2 != NOMARK && win[aw].winmrk2 >= tp)
        win[aw].winmrk2 += n1;
}

void
delewave(int num, int from, int to)
{
    int     c, i, nchn;
    long    n, npt, p1, p2;
    SIGPTR  pbuf1, pbuf2;
    SIGNAL *s;

    s = &sig[num];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == s->sigmrk[to])
	return;

    p1 = s->sigmrk[from];
    p2 = s->sigmrk[to];
    nchn = limit(1, s->sgnchn, MAX_CHAN);
    pbuf1 = s->sigbuf + p1 * nchn;
    pbuf2 = s->sigbuf + p2 * nchn;
    n = (s->siglen - p2) * nchn;
    while (n-- > 0)
	*pbuf1++ = *pbuf2++;

    npt = p2 - p1;
    for (c = 0; c < NMARK; c++) {
	if (s->sigmrk[c] != NOMARK && s->sigmrk[c] >= p1)
            s->sigmrk[c] = (s->sigmrk[c] < p2) ? NOMARK
	        : s->sigmrk[c] - npt;
    }
    s->siglen -= npt;
    if (curspnt > p1) {
	if (curspnt <= p2)
	    curspnt = p1;
	else
	    curspnt -= npt;
    }
    for (i = 0; i < nchn; i++)
        s->sigbuf[s->siglen * nchn + i] = s->sigzer;
    (void) sigsize(num, s->siglen, nchn);

    if (win[num].winmrk1 != NOMARK && win[num].winmrk1 >= p1)
        win[num].winmrk1 = (win[num].winmrk1 < p2) ? NOMARK
            : win[num].winmrk1 - npt;
    if (win[num].winmrk2 != NOMARK && win[num].winmrk2 >= p1)
        win[num].winmrk2 = (win[num].winmrk2 < p2) ? NOMARK
            : win[num].winmrk2 - npt;
}

void
sigdel(int num)
{
    (void) sigsize(num, sig[num].siglen = 0, 0);
    update_cursor();
    blankmarks(num);
    win[num].winmrk1 = win[num].winmrk2 = 0;
    showdata();
}

void
copy_rate(dst, src)
{
    sig[dst].adrate = sig[src].adrate;
    sig[dst].darate = sig[src].darate;
}

int
sigsize(num, size, nchan)
int     num, nchan;
long    size;
{
    long    i, n;
    SIGPTR  b, s;

    if (size < 1 || size > MAX_SAM)
	return (0);
    if (size < 1 || nchan < 1) {
	s = sig[num].sigbuf;
	if (s != (SIGPTR) NULL)
	    FREE(s);
	sig[num].sigbuf = (SIGPTR) NULL;
	return (1);
    }
    b = (SIGPTR) MALLOC((size + 1), sizeof(SIGPTR *) * nchan);
    if (b == (SIGPTR) NULL) {
	putmsgwk(" Memory allocation error. ");
    } else {
	s = sig[num].sigbuf;
	if (s != (SIGPTR) NULL) {
	    n = (sig[num].siglen + 1) * sig[num].sgnchn;
	    for (i = 0; i < n; i++)
		b[i] = s[i];
	    FREE(s);
	}
	sig[num].sigbuf = b;
	return (1);
    }
    return (0);
}

void
ordmrk()
{
    long    tmp;
    SIGNAL *s;

    if (seg_mark1 == NOMARK || seg_mark2 == NOMARK)
        return;
    s = &sig[aw];
    if (s->sigmrk[seg_mark1] > s->sigmrk[seg_mark2]) {
	tmp = seg_mark1;
	seg_mark1 = seg_mark2;
	seg_mark2 = tmp;
    }
}

/* nsrtzero - insert silence */

void
nsrtzero()
{
    char    str[80];
    double  dms;
    int     c, zero, i, nchn;
    long    cp, nsp, nzp, rate, n;
    SIGPTR  pbuf1, pbuf2;
    SIGNAL *s;
    static int def_bits = 16, def_zero = 0;

    s = &sig[aw];
    if (s->siglen == 0) {
	c = END_MARK;
    } else {
	putmsg(" INSERT SILENCE: in front of mark");
	c = getmark(0);
	clrmsg();
    }
    if (c == NOMARK)
	return;
    cp = sig[aw].sigmrk[c];
    if (cp == NOMARK)
	return;
    if (s->siglen == 0) {
	if (s->adrate <= 0)
	    s->adrate = def_rate;
	s->sgnchn = 1;
	s->sgbits = def_bits;
	s->sigzer = def_zero;
	s->sgtime = time(NULL);
	s->darate = s->adrate;
	s->sigbuf[s->siglen] = s->sigzer;
	sig->sigmrk[HOME_MARK] = 0;
	sig->sigmrk[END_MARK] = 0;
	sig->sigmrk[CURSOR_MARK] = curspnt = 0;
	blankmarks(aw);
	sprintf(str, "DURATION of silence (ms) = ");
    } else {
	sprintf(str, "DURATION of silence at mark %s (ms) = ", marktext(c));
    }
    get_response(str, str, 80);
    if (*str == '\0' || (dms = atof(str)) <= 0)
	return;
    rate = s->adrate;
    nchn = limit(1, s->sgnchn, MAX_CHAN);
    nsp = s->siglen;
    nzp = (int) (0.001 * dms * rate + 0.5);
    if (nzp < 0 || nzp > MAX_SAM - nsp) {
	putmsgwk(" Not enough room in signal buffer. ");
	beep();
	return;
    }
    if (sigsize(aw, nsp + nzp, nchn) == 0) {
	putmsgwk(" Can't expand the signal buffer. ");
	beep();
	return;
    }
    pbuf1 = s->sigbuf + (nsp - 1) * nchn;
    pbuf2 = s->sigbuf + (nsp - 1 + nzp) * nchn;
    n = nsp - cp;
    while (n-- > 0)
	*pbuf2-- = *pbuf1--;
    pbuf1 = s->sigbuf + cp * nchn;
    zero = s->sigzer;
    n = nzp * nchn;
    while (n-- > 0)
	*pbuf1++ = zero;

    s->siglen = nsp + nzp;
    for (c = 0; c < NMARK; c++)
	if (s->sigmrk[c] != NOMARK && s->sigmrk[c] >= cp)
	    s->sigmrk[c] += nzp;
    if (curspnt >= cp)
	curspnt += nzp;
    for (i = 0; i < nchn; i++)
        s->sigbuf[s->siglen * nchn + i] = s->sigzer;

    if (win[aw].winmrk1 != NOMARK && win[aw].winmrk1 == cp)
        win[aw].winmrk1 -= nzp;
    if (win[aw].winmrk2 != NOMARK && win[aw].winmrk2 >= cp)
        win[aw].winmrk2 += nzp;

    showdata();
}

/* blkxtnd - extend partial block with silence */

void
blkxtnd()
{
    int     nxt, zero, i, nchn;
    long    nsp, nzp, n;
    SIGPTR  pbuf;
    SIGNAL *s;

    s = &sig[aw];
    nchn = limit(1, s->sgnchn, MAX_CHAN);
    nsp = s->siglen;
    nxt = nsp % 256;
    if (nxt == 0)
	return;
    nzp = 256 - nxt;
    if (sigsize(aw, nsp + nzp, nchn) == 0) {
	putmsgwk(" Can't expand the signal buffer. ");
	beep();
	return;
    }
    pbuf = s->sigbuf + nsp * nchn;
    zero = s->sigzer;
    n = nzp;
    while (n-- > 0)
	*pbuf++ = zero;
    for (i = 0; i < nchn; i++)
        s->sigbuf[s->siglen * nchn + i] = zero;
    s->sigmrk[END_MARK] = s->siglen = nsp + nzp;
    win[aw].winmrk2 += nzp;
    showdata();
    sigmsg(aw);
}

void
editsamp()
{
    char    msg[80];
    int     nchn;

    if (sig[aw].siglen == 0)
        return;
    sprintf(msg, " Sample number [%ld] ?  ", curspnt);
    get_response(msg, msg, 80);
    if (*msg)
        curspnt = atol(msg);

    nchn = sig[aw].sgnchn;
    sprintf(msg, " Sample value [%d] ?  ", sig[aw].sigbuf[curspnt * nchn]);
    get_response(msg, msg, 80);
    if (*msg)
        sig[aw].sigbuf[curspnt * nchn] = atoi(msg);
    showdata();
    clrmsg();
}
