/* help.c */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "waved.h"
#include "video.h"
#include "fk.h"
#include "version.h"

#define ESC 27
#define NC  38
#define NR  21
#define TM  14

int     getevent();
int     ib_open();
void    ib_text();
void    ib_clear();
void    ib_close();

static int ibx1, iby1, ibx2, iby2, ibss;

static char *help_text_1[] = {
    VER,
    "                                      ",
    " [F1]     Help",
    " [F2]     Signal information",
    " [F3]     Toggle scroll-range box",
    " [F4]     Maximize window",
    " [F5]     Toggle scroll mode",
    "           (use < > to pan)",
    " [F6]     Toggle channel separation",
    " [Ins]    Insert a letter mark",
    " [Del]    Delete a letter mark",
    " [Esc]    Exit mode or program",
    " [Enter]  Replay signal to DAC",
    " [Tab]    Go to next window",
    " Ctrl-L   Redraw screen",
    "                                      ",
    "   <-     Cursor left",
    "   ->     Cursor right",
    " [Shift]  Cursor moves 10",
    " [Ctrl]   Cursor moves 50",
    "                                      ",
    "                                      ",
};
static char **help_text[] = {help_text_1};
static int np_help_text = sizeof(help_text) / sizeof(help_text[0]);

static char *sep[] =
{
    "",
    "   --------------------------------------------------------------  ",
    "",
};

static char *usage_text[] = {
    VER,
    "",
    "   usage:",
    "       waved [options] file ...",
    "",
    "   where 'file' is an ILS sampled data file.",
    "",
    "   options:",
    "",
    "       -H      write the help screen to file 'waved.hlp'",
    "       -g N    assume graphics adaptor N,",
    "               where N=cga,hgc,ega,vga",
    "       -m      monochrome (no colors)",
    "       -nh     assume no header on input files",
    "       -p N    assume I/O device N,",
    "               where N=dt,lpt,internal,sb,vmk",
    "       -r N    set default sample rate to N/sec",
    "",
};

/* helptext - write help screen to a file */

void
helptext()
{
    int     i, n;
    char   *name;
    FILE   *fp;

    name = "waved.hlp";
    fp = fopen(name, "w");
    if (fp == NULL) {
	fprintf(stderr, " can't open file '%s' \n", name);
	exit(1);
    }
    n = sizeof(usage_text) / sizeof(usage_text[0]);
    for (i = 0; i < n; i++)
	fprintf(fp, "%s\n", usage_text[i]);
    n = sizeof(sep) / sizeof(sep[0]);
    for (i = 0; i < n; i++)
	fprintf(fp, "%s\n", sep[i]);
    for (i = 0; i < NR - 1; i++)
	fprintf(fp, "    %s\n", help_text_1[i]);
    fclose(fp);
    fprintf(stderr, " WAVED help screen written to '%s' \n", name);
    exit(0);
}

void
showhelp()
{
    int     i, w, h, p;

    w = (NC + 2) * ch_wd;
    h = (NR + 1) * ln_sp;
    if (ib_open(w, h) < 0)
	return;
    p = 0;
    for (;;) {
	for (i = 0; i < NR; i++)
	    ib_text(ch_wd, i * ln_sp + TM, help_text[p][i]);
	if (getevent() != MMV)
	    break;
	p = (p + 1) % np_help_text;
        ib_clear();
    }
    ib_close();
}

void
usage()
{
    int     i, n;

    n = sizeof(usage_text) / sizeof(usage_text[0]);
    for (i = 0; i < n; i++)
	fprintf(stderr, "%s\n", usage_text[i]);
    exit(0);
}

int
ib_open(w, h)
{
    ibx1 = (scrszx - w) / 2;
    iby1 = (scrszy - h) / 2;
    ibx2 = ibx1 + w - 1;
    iby2 = iby1 + h - 1;
    ibss = g_savscr(ibx1, iby1, ibx2, iby2);
    if (ibss < 0) {
	putmsgwk("Problem opening info box.");
    } else {
        g_bar(ibx1, iby1, ibx2, iby2, hsbgc);
        g_box(ibx1 + 1, iby1 + 1, ibx2 - 1, iby2 - 1, hsfgc);
    }
    return (ibss);
}

void
ib_clear()
{
    g_bar(ibx1 + 2, iby1 + 2, ibx2 - 2, iby2 - 2, hsbgc);
}

void
ib_text(x, y, s)
char *s;
{
    g_settextclr(hsfgc, -1);
    g_text(x + ibx1, y + iby1, s);
}

void
ib_close()
{
    g_rstscr(ibss, ibx1, iby1);
}

