/* iod5.c - I/O Device = Turtle Beach Pinnacle/Fiji */

#include <dos.h>
#include "iod.h"

#define play_sig    play_pin
#define recrd_sig   recrd_pin

void stop_iod5();
int output_dac5(SIGPTR, long, long);
int input_adc5(SIGPTR, long, long);

int     chkevent();
int     getevent();
int     src_init(long, long, long, int, double, long *);
long    src_get(long, long, int, SIGPTR);
long    src_put(SIGPTR, long);
long    src_pull(long, long, int, SIGPTR *);
long    src_push(long, SIGPTR *);
void    msdelay(long);

static int play_sig();
static int recrd_sig();
static int init_dsp();
static int reset_dsp();
static void feed_sig();
static void read_cfg();

static int iod_present = 0;
static int iod_status;
static int srcflg = 0;
static long clk_rate = 0;
static long iod_npts, iod_rate = 0, iod_cnt;
static SIGPTR iod_buf;

int
init_iod5()
{
    adbits = 16;
    adzero = 0;
    if (!iod_present)
	read_cfg();
    if (init_dsp())
	return (1);
    feed_sig(adfeed);
    iod_present = 1;
    iod_rate = 0;
    stop_iod = stop_iod5;
    output_dac = output_dac5;
    input_adc = input_adc5;
    return (0);
}

void
stop_iod5()
{
    if (iod_present) {
	(void) reset_dsp();
        feed_sig(0);
    }
}

int
output_dac5(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    iod_buf = buf;
    iod_npts = npts;
    iod_rate = rate;
    iod_cnt = 0;
    iod_status = play_sig();
    return (iod_status);
}

int
input_adc5(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    iod_buf = buf;
    iod_npts = npts;
    iod_rate = rate;
    iod_cnt = 0;
    iod_status = recrd_sig();
    adnpts = iod_cnt;
    adrate = iod_rate;
    return (iod_status);
}

/************************************************************************/

#include<malloc.h>
#include<stdio.h>
#include<conio.h>

/*======================================================================*/

#define BUF_SIZE    0x4000
#define BNK_SIZE    (BUF_SIZE/2)

#define MSND_ICR    0x0		/* interrupt control register */
#define MSND_CVR    0x1		/* command vector */
#define MSND_ISR    0x2		/* interrupt status register */
#define MSND_IVR    0x3		/* interrupt vector register */
#define MSND_DSPR   0x4		/* DSP reset control register */
#define MSND_DAH    0x5		/* Tx/Rx data high */
#define MSND_DAM    0x6		/* Tx/Rx data middle */
#define MSND_DAL    0x7		/* Tx/Rx data low */

#define CMD_STOP	0x80 | 0x12
#define CMD_PLAY	0x80 | 0x13
#define CMD_RECORD	0x80 | 0x14
#define CMD_NOFEED	0x80 | 0x15
#define CMD_FEEDTHR	0x80 | 0x16
#define CMD_RATE	0x80 | 0x17
#define CMD_SETPOT	0x80 | 0x18
#define CMD_BNK1EMP	0x80 | 0x19
#define CMD_BNK1FUL	0x80 | 0x1A
#define CMD_BNK2EMP	0x80 | 0x1B
#define CMD_BNK2FUL	0x80 | 0x1C

static unsigned char pgm[0x600] = {
#include "pio.h"
};
static unsigned short cfgp = 0x260;
static unsigned short port = 0x290;
static unsigned short page = 0xD000;
static struct {int act, ad0, ad1, irq, ram;} res[5];

#ifdef _MSC_VER
#define MK_PTR(x,y)     _MK_FP(x,y)
#define ptr_enable()
#define ptr_disable()
#else
#include <sys/nearptr.h>
#define MK_PTR(x,y)     ((x)*16+(y)+__djgpp_conventional_base)
#define ptr_enable()    __djgpp_nearptr_enable();
#define ptr_disable()   __djgpp_nearptr_disable();
#endif /* _MSC_VER */
#define BNK_ADDR(b)     (SIGPTR)MK_PTR(page, BUF_SIZE * b)

static void
getbuf(SIGPTR sbuf, SIGPTR ibuf, long nw)
{
    int i;

    ptr_enable();
    for (i = 0; i < nw; i++)
        ibuf[i] = sbuf[i];
    ptr_disable();
}

static void
putbuf(SIGPTR sbuf, SIGPTR obuf, long nw)
{
    int i;

    ptr_enable();
    for (i = 0; i < nw; i++)
        sbuf[i] = obuf[i];
    ptr_disable();
}

static void
clrbuf(SIGPTR sbuf, long ip)
{
    int i;

    ptr_enable();
    for (i = ip; i < BNK_SIZE; i++)
        sbuf[i] = 0;
    ptr_disable();
}

static int
reset_dsp()
{
    int time_out = 20000;
    
    _outp(port + MSND_DSPR, 0); /* turn DSP reset on */
    _outp(port + MSND_DSPR, 2); /* turn DSP reset off */
    while(time_out-- >= 0) {
        if (_inp(port + MSND_CVR) == 0x12)
            return (1);
    }
    return (0);
}

static int
cvr_rdy()
{
    int time_out = 20000;
    
    while(time_out-- >= 0) {
        if ((_inp(port + MSND_CVR) & 0x80) == 0)
            return (1);
    }
    return (0);
}

static int
txd_rdy()
{
    int     time_out = 20000;

    while(time_out-- >= 0) {
	if (_inp(port + MSND_ISR) & 0x2)
	    return (1);
    }
    return (0);
}

static int
load_dsp_data(int w0, int w1, int w2)
{
    if (txd_rdy()) {
	_outp(port + MSND_DAH, w0);
	_outp(port + MSND_DAM, w1);
	_outp(port + MSND_DAL, w2);
	return (1);
    }
    return (0);
}

static void
dsp_cmd(int c)
{
    if (cvr_rdy())
	_outp(port + MSND_CVR, c);
}

static int
set_rate_code(int rc)
{
    if (load_dsp_data(0, 0, rc)) {
        dsp_cmd(CMD_RATE);
        return (0);
    }
    return (1);
}

static int
set_rate(long rate)
{
    int rc;

    if (rate == clk_rate)
        return (0);
    if (rate == 11025) {
	clk_rate = 11025;
	rc = 2;
    } else if (rate < 24000) {
	clk_rate = 22050;
	rc = 1;
    } else {
	clk_rate = 44100;
	rc = 0;
    }
    rc = set_rate_code(rc);
    msdelay(400L);
    return (rc);
}

static int
upload_pgm()
{
    unsigned char *bb, w0, w1, w2;
    int     i;
    static int nw = 512;

    bb = pgm;
    for (i = 0; i < nw; i++) {
	w0 = *bb++;
	w1 = *bb++;
	w2 = *bb++;
	if (!load_dsp_data(w0, w1, w2))
	    return (1);
    }
    return (0);
}

static int
rxd_rdy()
{
    return (_inp(port + MSND_ISR) & 0x1);
}

static void
get_dsp_data(int *w0, int *w1, int *w2)
{
    *w0 = _inp(port + MSND_DAH);
    *w1 = _inp(port + MSND_DAM);
    *w2 = _inp(port + MSND_DAL);
}

static void
feed_sig(int y)
{
    dsp_cmd(y ? CMD_FEEDTHR : CMD_NOFEED);
}

static void
set_sma()
{
    _outp(cfgp, 0x40);
    _outp(cfgp + 1, res[0].ram >> 8);
    _outp(cfgp, 0x41);
    _outp(cfgp + 1, res[0].ram & 0xFF);
}

static void
mem_blk_select(int p)
{
    _outp(port + MSND_DSPR, 2 | p);
}

static int
reset_sma()
{
    set_sma();
    mem_blk_select(0);
    clrbuf(BNK_ADDR(0), 0L);
    clrbuf(BNK_ADDR(1), 0L);
    return (0);	
}

static int 
set_pot_level(int left, int right, int select)
{
    if (load_dsp_data(left, right, select)) {
	dsp_cmd(CMD_SETPOT);
	return (0);
    }
    return (1);
}

static int
set_pots()
{
    static int inp = 255, aux = 0, mic = 255;

    if (set_pot_level(inp, inp, 0))
	return (1);
    if (set_pot_level(aux, aux, 1))
	return (2);
    if (set_pot_level(mic, mic, 2))
	return (3);
    return (0);
}

static int
init_dsp()
{
    int     w0, w1, w2;

    if (!reset_dsp())
	return (1);
    if (upload_pgm())
	return (2);
    while (rxd_rdy())
	get_dsp_data(&w0, &w1, &w2);
    if (reset_sma())
	return (3);
    if (set_pots())
	return (4);
    return (0);
}

static void
read_cfg()
{
    char line[80];
    int dev = 0;
    unsigned int v;
    FILE *cfp;
    static char *cfn1 = "pincfg.ini";
    static char *cfn2 = "c:/pinnacle/pincfg.ini";
    static char *cfn3 = "c:/pinnacle/program/pincfg.ini";
    
    if ((cfp = fopen(cfn1, "r")) == NULL)
        if ((cfp = fopen(cfn2, "r")) == NULL)
            if ((cfp = fopen(cfn3, "r")) == NULL)
                return;
    while(fgets(line, 80, cfp) != NULL) {
        if (strncmp(line, "[ConfigControl]", 15) == 0) {
            dev = -1;
        } else if (strncmp(line, "[LogicalDevice", 14) == 0) {
            dev = line[14] - '0';
        } else if (dev == -1 && strncmp(line, "IOAddress0", 10) == 0) {
            sscanf(line + 10, "=%x", &v);
            if (v >= 0x210 && v <= 0x3E0 && (v %0x10) == 0)
                cfgp = v;
        } else if (dev >= 0 && dev < 4) {
            if (strncmp(line, "Active", 6) == 0) {
                sscanf(line + 6, "=%x", &v);
                res[dev].act = v;
            } else if (strncmp(line, "IOAddress0", 10) == 0) {
                sscanf(line + 10, "=%x", &v);
                if (v >= 0x210 && v <= 0x3E0 && (v %0x10) == 0)
                    res[dev].ad0 = v;
            } else if (strncmp(line, "IOAddress1", 10) == 0) {
                sscanf(line + 10, "=%x", &v);
                if (v >= 0x210 && v <= 0x3E0 && (v %0x10) == 0)
                    res[dev].ad1 = v;
            } else if (strncmp(line, "IRQNumber", 9) == 0) {
                sscanf(line + 9, "=%x", &v);
                if (v >= 5 && v <= 12 && (v %0x10) == 0)
                    res[dev].ad0 = v;
            } else if (strncmp(line, "RAMAddress", 10) == 0) {
                sscanf(line + 10, "=%x", &v);
                if (v >= 0xB000 && v <= 0xE800 && (v %0x800) == 0)
                    res[dev].ram = v >> 4;
            }
        }
    }
    fclose(cfp);
    port = res[0].ad0;
    page = res[0].ram << 4;
}

static int bnkful[2];
static int curbnk = 0, curpnt = 0, iod_done = 0;

static int
play_buf(SIGPTR obuf, long nw, int eos)
{
    int     w0, w1, w2, bw;
    SIGPTR sbuf;
 
    if (!bnkful[curbnk] && nw > 0) {
        bw = BNK_SIZE - curpnt;
        if (nw > bw)
            nw = bw;
        sbuf = BNK_ADDR(curbnk);
        putbuf(&sbuf[curpnt], obuf, nw);
        curpnt += nw;
        if (eos || (curpnt == BNK_SIZE)) {
            if (nw < bw)
                clrbuf(sbuf, curpnt);
            dsp_cmd(curbnk == 0 ? CMD_BNK1FUL : CMD_BNK2FUL);
            dsp_cmd(CMD_PLAY);
            bnkful[curbnk] = 1;
            curbnk = 1 - curbnk;
            curpnt = 0;
        }
    } else {
    	nw = 0;
    }
    if ((!bnkful[0] && !bnkful[1]) || chkevent())
        dsp_cmd(CMD_STOP);
    if (rxd_rdy()) {
        get_dsp_data(&w0, &w1, &w2);
        if (w0 == 1) {
            bnkful[0] = 0;
        } else if (w0 == 3) {
            bnkful[1] = 0;
        } else {
            iod_done = 1;
        }
    }
    return (nw);
}

static int
play_sig()
{
    double gn = 0.95;
    int  eoso = 0, eosc = 0, fw = 16;
    long ipc, npc, ipo, npo, npp;
    SIGPTR cbf;

    if (set_rate(iod_rate))
    	return (1);
    srcflg = src_init(iod_npts, iod_rate, clk_rate, fw, gn, &npc);
    ipo = npo = ipc = 0;
    bnkful[0] = bnkful[1] = 0;
    iod_done = curpnt = curbnk = 0;
    iod_cnt = 0;
    while (!iod_done) {
        if (srcflg == 0) {
            iod_cnt += play_buf(&iod_buf[iod_cnt], iod_npts - iod_cnt, 1);
        } else {
            if (iod_cnt < iod_npts) {
                iod_cnt += src_put(&iod_buf[iod_cnt], iod_npts - iod_cnt);
            } else {
            	eosc = 1;
            }
            if (ipo >= npo) {
                npo = src_pull(ipc, npc - ipc, eosc, &cbf);
                if (npo == (npc - ipc))
                    eoso = 1;
                ipo = 0;
            }
            npp = play_buf(&cbf[ipo], npo, eoso);
            ipo += npp;
            ipc += npp;
        }
    }
    dsp_cmd(CMD_STOP);
    return (0);
}

static int
recrd_buf(SIGPTR ibuf, long nw)
{
    int     w0, w1, w2, bw;
    SIGPTR  sbuf;

    if (iod_cnt == iod_npts || chkevent())
        dsp_cmd(CMD_STOP);
    if (rxd_rdy()) {
        get_dsp_data(&w0, &w1, &w2);
        if (w0 == 2) {
            bnkful[0] = 1;
        } else if (w0 == 4) {
            bnkful[1] = 1;
        } else {
           iod_done = 1;
        }
    }
    if (bnkful[curbnk] && nw > 0) {
        bw = BNK_SIZE - curpnt;
        if (nw > bw)
            nw = bw;
        sbuf = BNK_ADDR(curbnk);
        getbuf(&sbuf[curpnt], ibuf, nw);
        curpnt += nw;
        if (curpnt == BNK_SIZE) {
            dsp_cmd(curbnk == 0 ? CMD_BNK1EMP : CMD_BNK2EMP);
            bnkful[curbnk] = 0;
            curbnk = 1 - curbnk;
            curpnt = 0;
        }
    } else {
        nw = 0;
    }
    return (nw);
}

static int
recrd_sig()
{
    double gn = 0.95;
    int  fw = 16;
    long ipc, npc;
    SIGPTR cbf;

    if (set_rate(iod_rate))
    	return (1);
    srcflg = src_init(iod_npts, clk_rate, iod_rate, fw, gn, &npc);
    npc = ipc = 0;
    bnkful[0] = bnkful[1] = 0;
    iod_done = curpnt = curbnk = 0;
    iod_cnt = 0;
    dsp_cmd(CMD_RECORD);
    while(!iod_done) {
        if (srcflg == 0) {
            iod_cnt += recrd_buf(&iod_buf[iod_cnt], iod_npts - iod_cnt);
        } else {
            if (ipc >= npc) {
                npc = src_push((long) BNK_SIZE, &cbf);
                ipc = 0;
            }
            ipc += recrd_buf(&cbf[ipc], npc - ipc);
            if (ipc >= npc) {
                iod_cnt += src_get(iod_cnt, iod_npts - iod_cnt, iod_done, 
                    &iod_buf[iod_cnt]);
            }
        }
    }
    dsp_cmd(CMD_STOP);
    return (0);
}
