/* iod3.c - I/O Device = Sound Blaster */

#include <dos.h>
#include "iod.h"

#define todc	(*((long FAR *) 0x46C))
#define CTPS 	18.207
#define CPS     1193180L
#define TIMER   8

#define DSP_VOIC_ON     0xd1
#define DSP_VOIC_OFF    0xd3

int     chkevent();
int     getevent();

static int sb_present();
static int sb_dsp_cmd();
static int sb_dsp_wr();
static int sb_dsp_rd();
static void delay4();
static void dsp_reset();
static void dsp_prime();
static void procspeed();

static void (INTERRUPT * oldtimer) ();
static void INTERRUPT dac_isr3();
static void INTERRUPT adc_isr3();
static SIGPTR isr_samp;
static long isr_count = 0;
static long isr_npts = 0;
static int isr_ticks = 0;
static int first_io = 1;
static int dly4 = 0;

int
init_iod3()
{
    adbits = 8;
    adzero = 128;
    if (dly4 == 0)
	procspeed();
    if (!sb_present())
	return (1);
    isr_count = 0;
    isr_npts = 0;
    if (first_io) {
	oldtimer = _dos_getvect(TIMER);
	first_io = 0;
    } else if (_dos_getvect(TIMER) != oldtimer) {
	set_timer_rate(0);
	_dos_setvect(TIMER, oldtimer);
    }
    return (0);
}

void
stop_iod3()
{
    if (isr_npts != 0) {
	isr_npts = 0;
	_dos_setvect(TIMER, oldtimer);
	set_timer_rate(0);
    }
    (void) sb_dsp_cmd(DSP_VOIC_OFF);
}

int
output_dac3(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    dsp_reset();
    (void) sb_dsp_cmd(DSP_VOIC_ON);
    dsp_prime(10);
    isr_ticks = CPS / rate;
    isr_samp = buf;
    isr_count = npts;
    isr_npts = npts;
    _dos_setvect(TIMER, dac_isr3);
    set_timer_rate(isr_ticks);
    while (isr_count > 0 && chkevent() == 0)
	continue;
    todc += (((isr_npts - isr_count) * isr_ticks) >> 16) + 1;
    set_timer_rate(0);
    _dos_setvect(TIMER, oldtimer);
    isr_npts = 0;
    (void) sb_dsp_cmd(DSP_VOIC_OFF);
    return (0);
}

int
input_adc3(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    dsp_reset();
    (void) sb_dsp_cmd(DSP_VOIC_OFF);
    (void) sb_dsp_rd();
    isr_ticks = CPS / rate;
    isr_samp = buf;
    isr_count = npts;
    isr_npts = npts;
    _dos_setvect(TIMER, adc_isr3);
    set_timer_rate(isr_ticks);
    while (isr_count > 0 && chkevent() == 0)
	continue;
    todc += (((isr_npts - isr_count) * isr_ticks) >> 16) + 1;
    set_timer_rate(0);
    _dos_setvect(TIMER, oldtimer);
    if (chkevent())
	(void) getevent();
    adnpts = isr_npts - isr_count;
    isr_npts = 0;
    return (0);
}

static void INTERRUPT
dac_isr3()
{
    if (isr_count > 0) {
	isr_count--;
	(void) sb_dsp_wr((int) *isr_samp++);
    } else {
	set_timer_rate(0);
    }
    (void) outp(0x20, 0x20);
}

static void INTERRUPT
adc_isr3()
{
    if (isr_count > 0) {
	isr_count--;
	*isr_samp++ = sb_dsp_rd();
    } else {
	set_timer_rate(0);
    }
    (void) outp(0x20, 0x20);
}

/****************************************************************/

#define CTPS 	        18.207
#define ADDR            0x388
#define DATA            0x389
#define FM_STAT         0x228
#define FM_DATA         0x229
#define DSP_RSET        0x226
#define DSP_RDAT        0x22A
#define DSP_WSTA        0x22C
#define DSP_WDAT        0x22C
#define DSP_RSTA        0x22E
#define DSP_WRIT_DIR    0x10
#define DSP_READ_DIR    0x20


void
procspeed()
{
    long    t;
    int     m, n, j, k = 100;

    for (dly4 = 1; dly4 < 256; dly4 *= 2) {
	for (t = todc + 1; todc < t;)
	    continue;
	t = todc;
	for (n = 0; todc == t; n++)
	    for (j = 0; j < 100; j++)
		delay4(1);
	if (n <= k)
	    break;
	m = n;
    }
    if (dly4 > 2)
	dly4 = (int) (((long) dly4 * (2 * m - (n + k))) / (2 * (m - n)));
}

void
delay4(int c)
{
    while (c-- > 0) {
	_asm mov cx, dly4 _asm dlup:loop dlup
    }
}

void
dsp_reset()
{
    outp(DSP_RSET, 1);
    delay4(20);
    outp(DSP_RSET, 0);
}

void
dsp_prime(n)
{
    while (n-- > 0)
	(void) sb_dsp_wr(adzero);
}

void
sb_reg_wr(int a, int d)
{
    outp(ADDR, a);
    delay4(1);			/* 3.3 us */
    outp(DATA, d);
    delay4(10);			/* 23 us */
}

int
sb_present()
{
    int     s0, s1;

    sb_reg_wr(4, 0x60);
    sb_reg_wr(4, 0x80);
    s0 = inp(FM_STAT) & 0xE0;
    sb_reg_wr(2, 0xFF);
    sb_reg_wr(4, 0x21);
    delay4(20);			/* at least 80 us */
    s1 = inp(FM_STAT) & 0xE0;
    sb_reg_wr(4, 0x60);
    sb_reg_wr(4, 0x80);
    return (s0 == 0x00 && s1 == 0xC0);
}

int
sb_dsp_cmd(int c)
{
    int     i;

    for (i = 0; i < 100; i++) {
	delay4(1);
	if ((inp(DSP_WSTA) & 0x80) == 0) {
	    outp(DSP_WDAT, c);
	    return (1);
	}
    }
    return (0);
}

int
sb_dsp_wr(int d)
{
    if (sb_dsp_cmd(DSP_WRIT_DIR)) {
	delay4(1);
	delay4(1);
	outp(DSP_WDAT, d);
	return (1);
    }
    return (0);
}

int
sb_dsp_rd()
{
    int     i;

    if (sb_dsp_cmd(DSP_READ_DIR)) {
	for (i = 0; i < 100; i++) {
	    delay4(1);
	    if (inp(DSP_RSTA) & 0x80) {
		delay4(1);
		return (inp(DSP_RDAT));
	    }
	}
    }
    return (adzero);
}
