/* iod_w32.c */

#include <windows.h>
#include <mmsystem.h>
#include <mmreg.h>
#include "waved.h"
#include "video.h"
#include "iod.h"

int     adbits = 8;
int     adzero = 128;
int     iod_num = -1;
long    adnpts = 0;

void
devicequery()
{
    putmsgwk(" This function is not implemented.");
}

void
find_iod()
{
    putmsgwk(" This function is not implemented.");
}

int
init_iod()
{
    adbits = 16;
    adzero = 0;
    iod_num = 0;
    return (0);
}

int
output_dac(SIGPTR buf, long npts, int nchn, long rate)
{
    HWAVEOUT        hWaveOut; 
    UINT            wResult; 
    WAVEFORMATEX    wfmt; 
    WAVEHDR         whdr; 

    // Check the event queue for early exit

    if (chkevent())
    	return (0);

    // Initialize the waveformat structure
    
    wfmt.wFormatTag = MM_MICROSOFT;
    wfmt.nChannels = nchn;
    wfmt.nSamplesPerSec = rate;
    wfmt.nAvgBytesPerSec = rate * 2;
    wfmt.nBlockAlign = 2;
    wfmt.wBitsPerSample = 16;
    wfmt.cbSize = 0;
    
    // Open a waveform device for output

    if (waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfmt, 0L, 0L, 
        CALLBACK_NULL | WAVE_ALLOWSYNC)) { 
        putmsgwk("Failed to open waveform output device.");
        return (0); 
    } 

    // Prepare wave_header. 
 
    whdr.lpData = (HPSTR) buf; 
    whdr.dwBufferLength = (DWORD) npts * nchn * 2; 
    whdr.dwFlags = 0L; 
    whdr.dwLoops = 0L; 
    waveOutPrepareHeader(hWaveOut, &whdr, sizeof(WAVEHDR)); 
 
    // Now the data buffer can be sent to the output device. The 
    // waveOutWrite function returns immediately and waveform 
    // data is sent to the output device in the background. 
 
    wResult = waveOutWrite(hWaveOut, &whdr, sizeof(WAVEHDR)); 
    if (wResult != 0) { 
        waveOutUnprepareHeader(hWaveOut, &whdr, sizeof(WAVEHDR)); 
        putmsgwk("Failed to write buffer to device");
        return (0); 
    } 

    // Wait for the wave_out to finish

    timer_wait(npts, rate);
    while (!(whdr.dwFlags & WHDR_DONE)) {
        if (chkevent()) {
            waveOutReset(hWaveOut);
            break;
        }
    }
    timer_kill();

    waveOutUnprepareHeader(hWaveOut, &whdr, sizeof(WAVEHDR)); 
    waveOutClose(hWaveOut);

    return (0);
}

int
input_adc(SIGPTR buf, long npts, long rate)
{
    HWAVEIN         hWaveIn; 
    UINT            wResult; 
    WAVEFORMATEX    wfmt; 
    WAVEHDR         whdr; 

    // Check the event queue for early exit

    if (chkevent())
    	return (0);

    // Initialize the waveformat structure
    
    wfmt.wFormatTag = MM_MICROSOFT;
    wfmt.nChannels = 1;
    wfmt.nSamplesPerSec = rate;
    wfmt.nAvgBytesPerSec = rate * 2;
    wfmt.nBlockAlign = 2;
    wfmt.wBitsPerSample = 16;
    wfmt.cbSize = 0;
    
    // Open a waveform device for input

    if (waveInOpen(&hWaveIn, WAVE_MAPPER, &wfmt, 0L, 0L, CALLBACK_NULL)) { 
        putmsgwk("Failed to open waveform input device.");
        return (0); 
    } 
 
    // Prepare wave_header. 
 
    whdr.lpData = (HPSTR) buf; 
    whdr.dwBufferLength = (DWORD) npts * 2; 
    whdr.dwFlags = 0L; 
    whdr.dwLoops = 0L; 
    waveInPrepareHeader(hWaveIn, &whdr, sizeof(WAVEHDR)); 
 
    // Now the data buffer can be added to the input device
    // and data acquired in the background. 
 
    wResult = waveInAddBuffer(hWaveIn, &whdr, sizeof(WAVEHDR)); 
    if (wResult != 0) { 
        waveInUnprepareHeader(hWaveIn, &whdr, sizeof(WAVEHDR)); 
        putmsgwk("Failed to add buffer to device");
        return (0); 
    } 
    waveInStart(hWaveIn);

    // Wait for the wave_in to finish
    
    timer_wait(npts, rate);
    while (!(whdr.dwFlags & WHDR_DONE)) {
        if (chkevent()) {
            waveInReset(hWaveIn);
            break;
        }
    }
    timer_kill();

    waveInUnprepareHeader(hWaveIn, &whdr, sizeof(WAVEHDR)); 
    waveInClose(hWaveIn);

    adnpts = whdr.dwBytesRecorded / 2;
    return (0);
}

void
stop_iod()
{
}
