/* lpc.c */

#define MAXNIT  1000

double  sqrt();

/* autoc - autocorrelation method of linear prediction analysis */

int
autoc(int n, float *x, int m, float *a, float *alpha, float *rc, float *r)
{
/*
 *   inputs:   n        - no. of data points
 *             x[n]     - input data sequence
 *             m        - order of filter
 *
 *  outputs:   a[m+1]   - filter coefficients
 *             alpha[1] - residual "energy"
 *             rc[m]    - reflection coefficients
 *             r[m+1]   - autocorrelation coefficients
 *
 */
    int     ib, ip, k;
    int     mh, minc, mp, nk, np;
    float   at, s, v;

    a[0] = 1.0;			/* first coefficient is always 1 */
    mp = m + 1;
    for (k = 0; k < mp; k++) {
	r[k] = 0.0;
	nk = n - k;
	for (np = 0; np < nk; np++) {
	    v = x[np] * x[np + k];
	    r[k] += v;
	}
    }
    if (r[0] <= 0.0)		/* check for zero energy in signal */
	return (1);
    rc[0] = -r[1] / r[0];
    a[1] = rc[0];
    alpha[0] = r[0] + r[1] * rc[0];
    for (minc = 2; minc <= m; minc++) {
	s = 0.0;
	for (ip = 0; ip < minc; ip++) {
	    s += r[minc - ip] * a[ip];
	}
	rc[minc - 1] = -s / alpha[0];
	mh = minc / 2 + 1;
	for (ip = 1; ip < mh; ip++) {
	    ib = minc - ip;
	    at = a[ip] + rc[minc - 1] * a[ib];
	    a[ib] += rc[minc - 1] * a[ip];
	    a[ip] = at;
	}
	a[minc] = rc[minc - 1];
	alpha[0] += rc[minc - 1] * s;
	if (alpha[0] <= 0.0)
	    return (1);
    }
    return (0);
}

int
streak(int xin, int m, float *rc)
{
    int     i;
    float   ep, em, den, emold, xk;

    ep = (float) xin;
    em = rc[0];
    rc[0] = ep;
    for (i = 1; i <= m; i++) {
	den = em * em + ep * ep;
	if (den == 0)
	    den = 1.0;
	xk = (float) (-2.0 * em * ep / den);
	emold = em;
	em = rc[i];
	rc[i] = xk * ep + emold;
	ep += xk * emold;
    }
    return ((int) ep);
}

int
findpk(f, npts, fsk, fm, bw)
float  *f, fsk, *fm, *bw;
int     npts;
{
    int     npeak, npts2, nptm1, nsl, nf, j;
    float   ts, tss, ym, yp, fn, fj, fjm1, fjp1, aa, bb, cc, pk, dd, gg;

    npeak = 0;
    npts2 = npts * 2;
    nptm1 = npts - 1;
    ts = (float) (fsk / npts2);
    tss = ts * ts;
    nf = 0;
    ym = f[1];
    yp = f[2];
    nsl = (yp < ym);
    for (j = 3; j < nptm1; j++) {
	ym = yp;
	yp = f[j];
	if (nsl && yp > ym) {
	    nsl = 0;
	} else if (!nsl && yp < ym) {
	    nsl = 1;
	    nf = nf + 1;
	    fn = (j - 1) * ts;
	    fjm1 = f[j];
	    fj = f[j - 1];
	    fjp1 = f[j];
	    aa = (fjm1 - 2 * fj + fjp1) / (2 * tss);
	    bb = (fjp1 - fjm1) / (2 * ts);
	    cc = fj;
	    pk = cc - (bb * bb) / (4 * aa);
	    dd = cc - pk / 2;
	    gg = bb * bb - 4 * aa * dd;
	    if (aa < 0.0 && gg >= 0.0) {
		bw[npeak] = (float) (-sqrt((double) gg) / aa);
		fm[npeak] = -bb / (2 * aa) + fn;
		npeak = npeak + 1;
	    }
	}
    }
    return (npeak);
}
