/* modify.c */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "waved.h"

char   *marktext();
int     getyes();
int     menu();

void    offseg();
void    scalseg();
void    revseg();
void    rmdcseg();
void    cmpseg();
void    invseg();
void    clipseg();
void    zerseg();
void    rescale();
void    reverse();
void    invert();
void    remove_dc();
void    compress();
void    centerclip();
void    zerowave();
void    chgbits();
void    offset();
void    addwin();
void    addwave();

void
modquery()
{
    int     c;
    static char *mis[] = {
        "Bits", "Compr", "Invert", "Ofst", "rmDC", "Scale", "cnTrclp",
        "Zero", "Reverse", "Add", ""
    };

    c = menu("MODIFY", mis);

    switch (c) {
    case 'C':
	cmpseg();
	break;
    case 'B':
	chgbits();
	break;
    case 'I':
	invseg();
	break;
    case 'O':
	offseg();
	break;
    case 'S':
	scalseg();
	break;
    case 'D':
	rmdcseg();
	break;
    case 'T':
	clipseg();
	break;
    case 'Z':
	zerseg();
	break;
    case 'R':
	revseg();
	break;
    case 'A':
	addwin();
	break;
    }
}

/* offseg - add offset to segment */

void
offseg()
{
    if (getwinmrk(" OFFSET:", 0)) {
	ordmrk();
	offset(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* scalseg - scale segment */

void
scalseg()
{
    if (getwinmrk(" RESCALE:", 0)) {
	ordmrk();
	rescale(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* revseg - reverse segment */

void
revseg()
{
    if (getwinmrk(" REVERSE:", 0)) {
	ordmrk();
	reverse(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* addwin - add another window */

void
addwin()
{
    addwave(clpwin);
    showdata();
    clrmsg();
}

/* rmdcseg - remove DC component from segment */

void
rmdcseg()
{
    if (getwinmrk(" REMOVE DC COMPONENT:", 0)) {
	ordmrk();
	remove_dc(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* cmpseg - compress segment */

void
cmpseg()
{
    if (getwinmrk(" COMPRESS:", 0)) {
	ordmrk();
	compress(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* invseg - invert segment */

void
invseg()
{
    if (getwinmrk(" INVERT:", 0)) {
	ordmrk();
	invert(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* clipseg - center clip segment */

void
clipseg()
{
    if (getwinmrk(" CENTER-CLIP:", 0)) {
	ordmrk();
	centerclip(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

/* zerseg - zero segment */

void
zerseg()
{
    int     c;

    if (getwinmrk(" ZERO:", 0)) {
	ordmrk();
	if (seg_mark1 == HOME_MARK && seg_mark2 == END_MARK) {
	    putmsg(" ZERO entire signal.  Are you sure? (Y/N) ");
	    c = getyes();
	    clrmsg();
	    if (!c)
		return;
	}
	zerowave(seg_mark1, seg_mark2);
	showdata();
    }
    clrmsg();
}

void
offset(int from, int to)
{
    char    str[80];
    double  val;
    int     nch;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    get_response("OFFSET: Enter the offset amount: ", str, 80);

    val = atof(str);
    if (val == 0.0)
	return;
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	buf[i] += (int) val;
    }
}

void
rescale(int from, int to)
{
    char    str[80];
    double  val;
    int     zer, nch;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    get_response("RESCALE: Enter the scale factor: ", str, 80);

    val = atof(str);
    if (val == 0.0)
	return;
    zer = s->sigzer;
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	buf[i] = (int) ((buf[i] - zer) * val + zer);
    }
}

void
reverse(int from, int to)
{
    int     buftmp, c, nch;
    long    aaa, bbb, i, j, npt, no2;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    no2 = npt / 2;
    for (i = 0; i < no2; i++) {
        j = npt - i;
	buftmp = buf[i];
	buf[i] = buf[j];
	buf[j] = buftmp;
    }
    for (c = 0; c < NMARK; c++) {
        if (s->sigmrk[c] >= aaa && s->sigmrk[c] < bbb)
            if (s->sigmrk[c] != NOMARK)
                s->sigmrk[c] = aaa + bbb - s->sigmrk[c];
    }

    if (win[aw].winmrk1 != NOMARK && win[aw].winmrk1 > aaa)
        win[aw].winmrk1 = NOMARK;
    if (win[aw].winmrk2 != NOMARK && win[aw].winmrk2 <= bbb)
        win[aw].winmrk2 = NOMARK;
}

void
invert(int from, int to)
{
    int     zer, nch;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    zer = s->sigzer;
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	buf[i] = zer - (buf[i] - zer);
    }
}

void
remove_dc(int from, int to)
{
    int     zer, dc, nch;
    double  sum, avg;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    sum = 0.0;
    zer = s->sigzer;
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	sum += (buf[i] - zer);
    }
    avg = sum / npt;
    dc = (int) ((avg >= 0) ? avg + 0.5 : avg - 0.5);
    for (i = 0; i < npt; i++) {
	buf[i] -= dc;
    }
}

void
compress(int from, int to)
{
    char    str[80];
    double  val, zer, tmp, scl;
    int     nch;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    get_response("COMPRESS: Enter the exponent: ", str, 80);

    val = atof(str);
    if (val < 0.0 || val > 1.0)
	return;
    zer = s->sigzer;
    scl = pow(2.0, (s->sgbits - 1) * (1.0 - val));
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	tmp = buf[i] - zer;
	if (tmp > 0.0)
	    buf[i] = (int) (zer + pow(tmp, val) * scl);
	else if (tmp < 0.0)
	    buf[i] = (int) (zer - pow(-tmp, val) * scl);
    }
}

void
centerclip(int from, int to)
{
    char    str[80];
    double  val;
    int     zer, tmp, nch;
    long    aaa, bbb, i, npt;
    SIGNAL *s;
    SIGPTR  buf;

    clrmsg();
    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];

    get_response("CENTER-CLIP: Enter the clip threshold: ", str, 80);

    val = atof(str);
    if (val == 0.0)
	return;
    zer = s->sigzer;
    nch = s->sgnchn;
    buf = s->sigbuf + aaa * nch;
    npt = (bbb - aaa) * nch;
    for (i = 0; i < npt; i++) {
	tmp = buf[i] - zer;
	if (tmp < val && tmp > -val)
	    buf[i] = zer;
    }
}

void
zerowave(int from, int to)
{
    int     zer, nch;
    long    p1, p2, npt;
    SIGNAL *s;
    SIGPTR  buf;

    s = &sig[aw];
    if (s == NULL)
	return;
    if (s->sigmrk[from] == NOMARK || s->sigmrk[to] == NOMARK)
	return;
    nch = s->sgnchn;
    p1 = s->sigmrk[from];
    p2 = s->sigmrk[to];
    if (p1 == p2) {
	return;
    } else if (p1 < p2) {
	buf = s->sigbuf + p1 * nch;
	npt = (p2 - p1) * nch;
    } else {
	buf = s->sigbuf + p2 * nch;
	npt = (p1 - p2) * nch;
    }
    zer = s->sigzer;
    while (npt-- > 0)
	*buf++ = zer;
}

void
chgbits()
{
    int     c, bits, zero;
    SIGNAL *s;
    static char *mis[] = {
        "A[16-bit,signed]", "B[12-bit,unsigned]", "C[8-bit,unsigned]", ""
    };
    c = menu("CHANGE BITS", mis);
    switch (c) {
    case 'A':
	bits = 16;
	zero = 0;
	break;
    case 'B':
	bits = 12;
	zero = 2048;
	break;
    case 'C':
	bits = 8;
	zero = 128;
	break;
    default:
	clrmsg();
	return;
    }
    s = &sig[aw];
    if (s->sgbits != bits || s->sigzer != zero)
	if (sigcon(aw, bits, zero, 0) && scrollmode == 0)
	    showdata();
    clrmsg();
}

void
addwave(num)
int     num;
{
    int     i, c, zer;
    long    n1, n2, ap, n;
    SIGNAL *s1, *s2;
    SIGPTR  pbuf1, pbuf2;

    if (num == aw)
	return;
    s1 = &sig[num];   /* clip window */
    s2 = &sig[aw];    /* active window */
    if (s1 == NULL || s2 == NULL)
	return;
    n1 = s1->siglen * s1->sgnchn;
    n2 = s2->siglen * s2->sgnchn;
    if (n1 == 0)
	return;
    putmsg(" ADD clip: at mark?");
    c = getmark(0);
    clrmsg();
    if (c == NOMARK)
	return;
    ap = s2->sigmrk[c];
    zer = s1->sigzer;
    pbuf1 = s1->sigbuf;
    pbuf2 = s2->sigbuf + ap * s2->sgnchn;
    n = (n1 < (n2 - ap)) ? n1 : (n2 - ap);
   for(i = 0; i < n; i++) {
	*pbuf2++ += (*pbuf1++ - zer);
    }
}
