/* mouse.c */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "waved.h"
#include "video.h"
#include "memory.h"
#include "fk.h"

#define TBSZ        13

void    window_max();
void    zoom_in();
void    zoom_out();
void    pan_left();
void    pan_right();
void    pan_left_sm();
void    pan_right_sm();
void    replay();
void    window_edit();
void    window_scroll();

void    set_mouse_event();
void    get_mouse_location();

#ifndef WIN32
int     g_minit();
void    g_mshow();
static int mouse_sh = 0;
static int mouse_on = 0;
#endif /* WIN32 */

void
mouse_init()
{
#ifndef WIN32
    if (g_minit()) {
	mouse_on = 1;
	mouse_sh = 0;
	g_mshow(mouse_sh);
	set_mouse_event(1);
    }
#endif /* WIN32 */
}

void
mouse_shows(m)
{
#ifndef WIN32
    if (mouse_on)
	g_mshow(mouse_sh = m);
#endif /* WIN32 */
}

int
inwind(int x, int y, int wn)
{
    WINDOW *w;

    w = &win[wn];
    return (y >= w->winy1 && y <= w->winy2);
}

int
mouse_aw()
{
    int     x, y, s;
    long    n, d, p;
    WINDOW *w;

    get_mouse_location(&x, &y);
    w = &win[aw];
    if (y < w->winy1 + 15 || scrollmode) {
	s = 1;
    } else {
	n = w->winmrk2 - w->winmrk1;
	d = w->winx2 - w->winx1 - 4;
	p = w->winmrk1 + ((x - 2) * n + (d / 2)) / d;
	if (p <= 0) {
            s = 2;
	} else if (p < w->winmrk1) {
            s = 3;
	} else if (p >= sig[aw].siglen) {
            s = 4;
	} else if (p > w->winmrk2) {
            s = 5;
	} else {
            s = 6;
	    curspnt = p;
	}
    }
    return (s);
}

void
toolbar()
{
    int tbp = scrszx - TBSZ * ch_wd;

    g_settextclr(tlbcol, msbgc);
    if (scrollmode)
        g_text(tbp, stsy, "     < > \" ~ ");
    else
        g_text(tbp, stsy, " + - < > \" ~ ");
}

void
chk_toolbar(x, y)
{
    int tbp = scrszx - TBSZ * ch_wd;
    int i, xx, x0, y0, x1 = 0, x2;
    static void (*tool[])() = {
    	zoom_in, zoom_out, pan_left, pan_right, replay, window_scroll
    };
    static int nt = sizeof(tool) / sizeof(tool[0]);
    
    x0 = tbp;
    y0 = stsy - ch_ht;
    if ((x >= tbp) && (y >= y0) && (y <= (y0 + ln_sp))) {
        x1 = (ch_wd * 3) / 4;
        x2 = (ch_wd * 9) / 4;
        for (i = 0; i < nt; i++) {
            xx = x - (i * ch_wd * 2 + x0);
            if (xx >= x1 && xx < x2)
                tool[i]();
        }
    }
}

void
chk_toolbar_sm(x, y)
{
    int tbp = scrszx - TBSZ * ch_wd;
    int i, xx, x0, y0, x1 = 0, x2;
    static void (*tool[])() = {
    	pan_left_sm, pan_right_sm, replay, window_edit
    };
    static int nt = sizeof(tool) / sizeof(tool[0]);
    
    x0 = tbp + ch_wd * 4;
    y0 = stsy - ch_ht;
    if (x >= tbp && y >= y0 && y <= (y0 + ln_sp))
        x1 = (ch_wd * 3) / 4;
        x2 = (ch_wd * 9) / 4;
        for (i = 0; i < nt; i++) {
            xx = x - (i * ch_wd * 2 + x0);
            if (xx >= x1 && xx < x2)
                tool[i]();
        }
}

void
mouse_left()
{
    int     i, x, y, c;

    get_mouse_location(&x, &y);
    
    for (i = fsw; i <= lsw; i++) {
	if (inwind(x, y, i)) {
	    if (i == aw) {
                c = mouse_aw();
		if (c == 1)
		    window_max();
		else if (c == 2)
                    goto_samp(sig[aw].sigmrk[HOME_MARK]);
		else if (c == 3)
                    pan_left();
		else if (c == 4)
                    goto_samp(sig[aw].sigmrk[END_MARK]);
		else if (c == 5)
                    pan_right();
	    } else if (!maximize) {
		chaw(i);
            }
	}
    }
    chk_toolbar(x, y);
}

void
mouse_right()
{
    int     x, y;
    long    n, d, p;
    WINDOW *w;

    get_mouse_location(&x, &y);
    if (inwind(x, y, aw)) {
        if (mouse_aw() == 1) {
            blankmarks(aw);
            showdata();
        } else {
            w = &win[aw];
            n = w->winmrk2 - w->winmrk1;
            d = w->winx2 - w->winx1 - 4;
            p = w->winmrk1 + ((x - 2) * n + (d / 2)) / d;
            if (p >= w->winmrk1 && p <= w->winmrk2) {
                curspnt = p;
                ins_new_mark();
            }
	}
    } else {
	replay();
    }
}

int
menu(char *prmpt, char **mis)
{
    char    msg[81], *s;
    int     x1[20], x2[20], mic[20];
    int     i, j, c, n, m, nmi, mcx, mcy;

    sprintf(msg, " %s: ", prmpt);
    for (i = 0; mis[i] && *mis[i] && i < 20; i++) {
	s = mis[i];
	m = strlen(s);
	n = strlen(msg);
	if (n + m > 80)
	    break;
	sprintf(msg + n, " %s ", s);
	x1[i] = (n + 1) * 8;
	x2[i] = (n + 1 + m) * 8;
	for (j = 0; j < m; j++) {
	    c = s[j];
	    if ((c >= 'A' && c <= 'Z') || (c >= 0 && c <= 9)) {
		mic[i] = c;
		break;
	    }
	}
    }
    nmi = i;

/* getevent */

    if (!chkevent()) {
	putmsg(msg);
	toolbar();
    }
    c = getevent();
    if (c == MDL) {
	get_mouse_location(&mcx, &mcy);
	if (mcy >= (msgy - 8) && mcy <= msgy)
	    for (i = 0; i < nmi; i++)
		if (mcx >= x1[i] && mcx <= x2[i])
		    return (mic[i]);
    }
    if ((c & 0xFF00) == 0)
	c = toupper(c);
    return (c);
}

void
set_mmv(int status)
{
#ifndef WIN32
    if (mouse_on)
	set_mouse_event(status ? 2 : 1);
#endif /* WIN32 */
}
