; pio.asm - record and play routines for TB Pinnacle/Fiji soundcard
;
; I/O register equates
;
PBCR    EQU  $FFE0    ; port B control register
PCCR    EQU  $FFE1    ; port C control register
PCDD    EQU  $FFE3    ; port C data direction register
HOCR    EQU  $FFE8    ; host control register
HOSR    EQU  $FFE9    ; host status register
HOTR    EQU  $FFEB    ; host receive data register
SSCA    EQU  $FFEC    ; SSI control register A
SSCB    EQU  $FFED    ; SSI control register B
SSSR    EQU  $FFEE    ; SSI status/time slot register
STRD    EQU  $FFEF    ; serial transmit/receive data register
PABC    EQU  $FFFE    ; port A bus control register
IPRI    EQU  $FFFF    ; interrupt priority register
PLLC    EQU  $FFFD    ; PLL control register
SRCR    EQU  $FFC0    ; sample rate control register
CLKE    EQU  $FFC2    ; sample clock enable ???
VOLC    EQU  $FFC4    ; volume control register
;
;
        ORG P:$0000
; interrupt vector [00] - hardware reset
        JMP <START                              ; RESET
;
        ORG P:$0024
; interrupt vector [12] - host command 'STOP'
        BCLR #$02,Y:<$04                        ; clear play bit
        BCLR #$03,Y:<$04                        ; clear record bit
; interrupt vector [13] - host command 'PLAY'
        BSET #$02,Y:<$04                        ; set play bit
        NOP                                     ; [place holder]
; interrupt vector [14] - host command 'RECORD'
        BSET #$03,Y:<$04                        ; set record bit
        NOP                                     ; [place holder]
; interrupt vector [15] - host command 'NOFEED'
        BCLR #$01,Y:<$04                        ; clear feedthru bit
        NOP                                     ; [place holder]
; interrupt vector [16] - host command 'FEEDTHR'
        BSET #$01,Y:<$04                        ; P:002C 0A0461
        NOP                                     ; [place holder]
; interrupt vector [17] - host command 'RATE'
        JSR >SETRATE                            ; call SETRATE routine
; interrupt vector [18] - host command 'SETPOT'
        JSR >SETPOT                             ; call SETPOT routine
; interrupt vector [19] - host command 'BNK1EMP'
        BCLR #$0D,Y:<$04                        ; clear bank 1 bit
        NOP                                     ; [place holder]
; interrupt vector [1A] - host command 'BNK1FUL'
        BSET #$0D,Y:<$04                        ; set bank 1 bit
        NOP                                     ; [place holder]
; interrupt vector [1B] - host command 'BNK2EMP'
        BCLR #$0F,Y:<$04                        ; clear bank 2 bit
        NOP                                     ; [place holder]
; interrupt vector [1C] - host command 'BNK2FUL'
        BSET #$0F,Y:<$04                        ; set bank 2 bit
        NOP                                     ; [place holder]
;
        ORG P:$0040
;
; interrupt routine [00] - hardware reset
;
START   ORI #$03,MR                             ; mask interrupts
        CLR A                                   ; put zero in reg A
        MOVEP #>$0004,X:<<PABC                  ; set mem wait states
        MOVEP #>$0001,X:<<PBCR                  ; Port-B set periph
        MOVEP #>$0004,X:<<HOCR                  ; Host cmd interrupt enable
        MOVEP A,X:<<PCCR                        ; Port-C clear control reg
        MOVEP #>$0008,X:<<PCDD                  ; Port-C cfg pin3 output
        MOVEP #>$FFFFE7,X:<<PCCR                ; Port-C periph except pin3,4
        MOVEP A,X:<<STRD                        ; SSI clear data reg
        MOVEP #>$6100,X:<<SSCA                  ; SSI 2wrd/frm, 24bit/wrd
        MOVEP #>$FA00,X:<<SSCB                  ; SSI T/R enbl, net.mode
                                                ; cont.clk
        MOVEP #>$0400,X:<<IPRI                  ; IntPrior HST=1 SCI=0 SSI=0
        MOVEP A,X:<<STRD                        ; SSI clear data reg
        MOVE                     A,Y:<$04       ; clear PLAY/RECORD flags
        MOVEP #>$150003,X:<<PLLC
        MOVEP #$B,Y:<<SRCR                      ; set sample clock rate
        MOVEC  A,SR                             ; enable interrupts
IDLUP
        JSSET #$02,Y:<$04,>PLAY
        JSSET #$03,Y:<$04,>RECORD
        JCLR #$01,Y:<$04,>IDLUP
WAD0    JCLR #$07,X:<<SSSR,>WAD0
        MOVEP X:<<STRD,B
        MOVE                     B,X:>STRD
        JMP <IDLUP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PLAY

SYNC0   MOVEP #>$0000,X:<<STRD                  ; output zero to DAC
SYNC1   JCLR #$06,X:<<SSSR,>SYNC1               ; wait for txd empty
        JSET #$02,X:<<SSSR,>SYNC0               ; loop if no frame-sync occurred
SYNC2   MOVEP #>$0000,X:<<STRD                  ; output zero to DAC
SYNC3   JCLR #$06,X:<<SSSR,>SYNC3               ; wait for txd empty
        JSET #$02,X:<<SSSR,>SYNC2               ; loop if frame-sync occurred

PCKB1   JCLR #$02,Y:<$04,>PLAYX
        JSET #$0D,Y:<$04,>PLBF1
        JSR >ZERDA
        JMP <PCKB2
PLBF1   MOVE                     #>$4000,R0     ; load buffer 1 start addr
        JSR >PLBUF                              ; play buffer 1
        BCLR #$0D,Y:<$04                        ; clear buffer 1 full flag
PWHO1   JCLR #$01,X:<<HOSR,>PWHO1               ; wait for host reg
        MOVEP #>$10000,X:<<HOTR                 ; send buffer 1 code to host
PCKB2   JCLR #$02,Y:<$04,>PLAYX
        JSET #$0F,Y:<$04,>PLBF2
        JSR >ZERDA
        JMP <PCKB1
PLBF2   MOVE                     #>$6000,R0     ; load buffer 2 start addr
        JSR >PLBUF                              ; play buffer 2
        BCLR #$0F,Y:<$04                        ; clear buffer 2 full flag
PWHO2   JCLR #$01,X:<<HOSR,>PWHO2               ; wait for host reg
        MOVEP #>$30000,X:<<HOTR                 ; send buffer 2 code to host
        JMP <PCKB1
PLAYX   JSR >ZERDA
PWHOX   JCLR #$01,X:<<HOSR,>PWHOX               ; wait for host reg
        MOVEP #>$50000,X:<<HOTR                 ; send exit code to host
        RTS                                     ; return from PLAY
;
PLBUF   MOVE                     #>$2000,X0     ; play 1 buffer
        DO X0,>P00B4                            ; loop over npts
P00AE   JCLR #$06,X:<<SSSR,>P00AE               ; wait for DAC1
        MOVEP X:(R0),X:<<STRD                   ; output DAC1
P00B1   JCLR #$06,X:<<SSSR,>P00B1               ; wait for DAC2
        MOVEP X:(R0)+,X:<<STRD                  ; output DAC2
P00B4
        RTS
;
ZERDA   DO #$002,>ZRDA1                         ; zero both DACs
ZRDA0   JCLR #$06,X:<<SSSR,>ZRDA0               ; wait for DAC
        MOVEP #>$0000,X:<<STRD                  ; output zero
ZRDA1
        RTS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
RECORD

        MOVE                     X:>SSSR,X0     ; read SSI status reg
RSYNC   MOVE                     X:>STRD,X0     ; read SSI data reg
RWAD0   JCLR #$07,X:<<SSSR,>RWAD0               ; wait for ADC
        JCLR #$03,X:<<SSSR,>RSYNC               ; wait for frame flag
;
        BCLR #$0D,Y:<$04                        ; clear buffer 1 full flag
        BCLR #$0F,Y:<$04                        ; clear buffer 2 full flag
;
RCKB1   JCLR #$03,Y:<$04,>RECXT                 ; check RECORD flag
        JCLR #$0D,Y:<$04,>RECB1                 ; check buffer 1 flag
        JSR >ZERHO
        JMP <RCKB2
RECB1   MOVE                     #>$4000,R0     ; load buffer 1 start addr
        JSR >RECBF                              ; record buffer 1
        BSET #$0D,Y:<$04                        ; set buffer 1 full flag
;
RWHO1   JCLR #$01,X:<<HOSR,>RWHO1               ; wait for host reg
        MOVEP #>$20000,X:<<HOTR                 ; send buffer 2 code to host
RCKB2   JCLR #$03,Y:<$04,>RECXT                 ; check RECORD flag
        JCLR #$0F,Y:<$04,>RECB2                 ; check buffer 2 flag
        JSR >ZERHO
        JMP <RCKB1
RECB2   MOVE                     #>$6000,R0     ; load buffer 2 start addr
        JSR >RECBF                              ; record buffer 2
P00D8   JCLR #$01,X:<<HOSR,>P00D8               ; wait for host reg
        MOVEP #>$40000,X:<<HOTR                 ; send buffer 2 code to host
        BSET #$0F,Y:<$04                        ; set buffer 2 full flag
        JMP <RCKB1
RECXT   
RWHOX   JCLR #$01,X:<<HOSR,>RWHOX               ; wait for host reg
        MOVEP #>$60000,X:<<HOTR                 ; send exit code to host
        RTS
;
RECBF   MOVE                     #>$2000,X0     ; record buffer
        DO X0,>RCBF3                            ; loop over npts
RCBF1   JCLR #$07,X:<<SSSR,>RCBF1               ; wait for ADC
        MOVEP X:<<STRD,A                        ; put DAC value in reg A
        MOVE                     A,X:(R0)+      ; put DAC value in buffer
        JCLR #$01,Y:<$04,>RCBF2                 ; check feedthru flag
        MOVE                     A,X:>STRD      ; output DAC value to ADC
RCBF2   JCLR #$07,X:<<SSSR,>RCBF2               ; wait for ADC
        MOVEP X:<<STRD,A                        ; read DAC and ignore
RCBF3
        RTS                                     ; return from RECBF
;
ZERHO   JCLR #$01,X:<<HOSR,>ZERHO               ; wait for host
        MOVEP #>$0000,X:<<HOTR                  ; send zero to host
        RTS                                     ; return
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; interrupt routine [17] - host command 'RATE'
;
SETRATE
        JCLR #$00,X:<<HOSR,>SETRATE             ; wait for host data
        MOVEP X:<<HOTR,B                        ; read host data into reg B
        BSET #$00,Y:<<CLKE                      ; enable sample clock
        MOVE                     #>$000B,A      ; put #11 into reg A
        MOVE                     #>$0003,Y0     ; mask lower 2 bits
        AND Y0,B                                ; get lower 2 bits of reg B
        SUB B,A                                 ; subtract B from A
        MOVEP A,Y:<<SRCR                        ; set sample clock rate
        RTI                                     ; return from SETRATE
;
; interrupt routine [18] - host command 'SETPOT'
;
SETPOT   
        JCLR #$00,X:<<HOSR,>SETPOT              ; wait for host data
        MOVEP X:<<HOTR,A1                       ; read host data into reg A1
                                                ; A1=LLRRPP: left,right,POT
        JCLR #$00,A1,>EVEN                      ; check pot-select bit 0
        JCLR #$01,A1,>POT1                      ; select POT1
POT3    JMP <SPOT3                              ; select POT3
POT2    MOVE                     #>$008B,X0     ; POT2 & AMPS bits
        JMP <SPOT1
EVEN    JCLR #$01,A1,>POT0
POT1    MOVE                     #>$008D,X0     ; POT1 & AMPS bits
        JMP <SPOT1
POT0    MOVE                     #>$008E,X0     ; POT0 & AMPS bits
SPOT1   MOVEP X0,Y:<<VOLC                       ; enable AMPS and selected pot
        DO #$010,>SPOT3                         ; loop over LLRR bits
        ROL A                                   ; rotate out leading bit
        JCC >SPOT2                              ; if CARRY=0, then pot-bit=0
        BSET #$05,X0                            ; otherwise vol-bit=1
SPOT2   MOVEP X0,Y:<<VOLC                       ; output vol bit w/ lo clock 
        BSET #$06,X0                            ; set clock bit
        MOVEP X0,Y:<<VOLC                       ; output vol bit w/ hi clock
        BCLR #$05,X0                            ; clear clock bit
        BCLR #$06,X0                            ; clear vol bit
        MOVEP X0,Y:<<VOLC                       ; output lo vol & clock bits
SPOT3
        MOVEP #>$0080,Y:<<VOLC                  ; enable AMPS
        RTI                                     ; return from SETPOT
