/* rfft.c */

#include <math.h>

static int bitrev();
static int n2pow();
static void rfft2();
static void rfft4();
static void rfftbr();

static double tpi = 2.0 * 3.1415926535898;
static double srh = 0.70710678118655;

int
rfft(w, n)
float  *w;
int     n;
{
    int     i, m, nn, n4pow;

    if ((m = n2pow(n)) <= 0)
	return (1);
    n4pow = m / 2;
    if (m % 2 == 0) {
	nn = 1;
    } else {
	nn = 2;
	rfft2(n / nn, w);
    }
    for (i = 0; i < n4pow; i++) {
	nn *= 4;
	rfft4(n / nn, nn, w);
    }
    rfftbr(m, w);
    return (0);
}

static void
rfft2(i, w)
int     i;
float  *w;
{
    int     k;
    float   t;
    float  *w0, *w1;

    w0 = w;
    w1 = w0 + i;
    for (k = 0; k < i; k++) {
	t = w0[k] + w1[k];
	w1[k] = w0[k] - w1[k];
	w0[k] = t;
    }
}

static void
rfft4(i, nn, w)
int     i, nn;
float  *w;
{
    int     i4, j, jo, ji, jj, jl, jr, jt, k, ko, kl, m, n, ni;
    float   arg, c1, c2, c3, s1, s2, s3, vr, vi, r1, r5;
    float   t0, t1, t2, t3, t4, t5, t6, t7, tpiovn;
    float  *w0, *w1, *w2, *w3, *w4, *w5, *w6, *w7;

    w4 = w0 = w;
    w5 = w1 = w0 + i;
    w6 = w2 = w1 + i;
    w7 = w3 = w2 + i;
    n = nn / 4;
    m = n2pow(n);
    tpiovn = (float) (tpi / nn);
    ji = 3;
    jl = 2;
    jr = 2;
    ni = (n + 1) / 2;
    for (jj = 0; jj < ni; jj++) {
	if (jj == 0) {
	    for (k = 0; k < i; k++) {
		t0 = w0[k] + w2[k];
		t1 = w1[k] + w3[k];
		w2[k] = w0[k] - w2[k];
		w3[k] = w1[k] - w3[k];
		w0[k] = t0 + t1;
		w1[k] = t0 - t1;
	    }
	    if (nn > 4) {
		ko = i * 4;
		kl = ko + i;
		for (k = ko; k < kl; k++) {
		    vr = (float) (srh * (w1[k] - w3[k]));
		    vi = (float) (srh * (w1[k] + w3[k]));
		    w3[k] = w2[k] + vi;
		    w1[k] = vi - w2[k];
		    w2[k] = w0[k] - vr;
		    w0[k] += vr;
		}
	    }
	} else {
	    arg = tpiovn * bitrev(jj, m);
	    c1 = (float) cos(arg);
	    s1 = (float) sin(arg);
	    c2 = c1 * c1 - s1 * s1;
	    s2 = c1 * s1 + c1 * s1;
	    c3 = c1 * c2 - s1 * s2;
	    s3 = c2 * s1 + s2 * c1;
	    i4 = i * 4;
	    jo = jr * i4;
	    ko = ji * i4;
	    jt = jo + i;
	    for (j = jo; j < jt; j++) {
		k = ko + j - jo;
		r1 = w1[j] * c1 - w5[k] * s1;
		r5 = w1[j] * s1 + w5[k] * c1;
		t2 = w2[j] * c2 - w6[k] * s2;
		t6 = w2[j] * s2 + w6[k] * c2;
		t3 = w3[j] * c3 - w7[k] * s3;
		t7 = w3[j] * s3 + w7[k] * c3;
		t0 = w0[j] + t2;
		t4 = w4[k] + t6;
		t2 = w0[j] - t2;
		t6 = w4[k] - t6;
		t1 = r1 + t3;
		t5 = r5 + t7;
		t3 = r1 - t3;
		t7 = r5 - t7;
		w0[j] = t0 + t1;
		w7[k] = t4 + t5;
		w6[k] = t0 - t1;
		w1[j] = t5 - t4;
		w2[j] = t2 - t7;
		w5[k] = t6 + t3;
		w4[k] = t2 + t7;
		w3[j] = t3 - t6;
	    }
	    jr += 2;
	    ji -= 2;
	    if (ji <= jl) {
		ji = 2 * jr - 1;
		jl = jr;
	    }
	}
    }
}

static void
rfftbr(m, w)
int     m;
float  *w;
{
    int     j, k, kl, n;
    int     ji, ij, i;
    float   t;

    k = 4;
    kl = 2;
    n = 1 << m;
    for (j = 4; j <= n; j += 2) {
	if (k > j) {
	    t = w[j - 1];
	    w[j - 1] = w[k - 1];
	    w[k - 1] = t;
	}
	k -= 2;
	if (k <= kl) {
	    k = 2 * j;
	    kl = j;
	}
    }
    for (ij = 0; ij <= (n - 2); ij += 2) {
	ji = bitrev(ij >> 1, m) << 1;
	if (ij < ji) {
	    t = w[ij];
	    w[ij] = w[ji];
	    w[ji] = t;
	    t = w[ij + 1];
	    w[ij + 1] = w[ji + 1];
	    w[ji + 1] = t;
	}
    }
    for (i = 3; i < n; i += 2)
	w[i] = -w[i];
}

static int
bitrev(i, m)
register int i, m;
{
    register int j;

    j = i & 1;
    --m;
    while (--m > 0) {
	i >>= 1;
	j <<= 1;
	j |= i & 1;
    }
    return (j);
}

static int
n2pow(n)
{
    int     m;
    static int max = sizeof(int) * 8;

    for (m = 1; m < max; m++)
	if (n == (1 << m))
	    return (m);
    return (-1);
}
