/* savefile.c */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
#include "waved.h"
#include "video.h"
#include "file.h"
#ifdef DJC
#include <unistd.h>
#endif

#ifdef _O_BINARY
#define O_MODE      _O_WRONLY | _O_CREAT | _O_TRUNC| _O_BINARY
#else
#define O_MODE      O_WRONLY | O_CREAT | O_TRUNC
#endif /* _O_BINARY */
#define P_MODE      S_IREAD | S_IWRITE
#ifndef _open
#define _access     access
#define _close      close
#define _open       open
#define _lseek      lseek
#define _write      write
#define _swab       swab
#endif /* _open */

void    get_file_name();
void    exch_snd_hdr(SND_fmt *);
extern char *ft[NFT];

static int wbopen();
static void savefile_c();
static void savefile_plt();
static void savefile_lst();
static void savefile_ils();
static void savefile_mat();
static void savefile_wav();
static void savefile_noh();
static void savefile_snd();
static void savefile_adf();
static void write_data();
static void remext();
static void defext();
static FILE *wtopen();

void
savefile(aaa, bbb)
long    aaa, bbb;
{
    char    name[48], str[80];

    if (bbb <= aaa) {
	putmsgwk("No signal.");
	return;
    }
    strncpy(name, sig[aw].sigfnm, 48);
    if (*name <= ' ' || *name == '[') {
	*name = '\0';
    } else if (sig[aw].sigfty != file_type || strchr(name, ',') != NULL) {
	remext(name);
        defext(name);
    }
    sprintf(str, "Enter name of (%s) output file:  ", ft[file_type]);
    get_file_name(str, name, FNSZ);

    if (strlen(name) == 0)
	return;
    if (_access(name, 0) == 0) {
	sprintf(str, " Overwrite existing file '%s' (y/N) ? ", name);
	putmsg(str);
	if (!getyes())
	    return;
    }
    switch (file_type) {
    case FT_C:
	savefile_c(name, aaa, bbb);
	break;
    case FT_LST:
	savefile_lst(name, aaa, bbb);
	break;
    case FT_ILS:
	savefile_ils(name, aaa, bbb);
	break;
    case FT_PLT:
	savefile_plt(name, aaa, bbb);
	break;
    case FT_SND:
	savefile_snd(name, aaa, bbb);
	break;
    case FT_MAT:
	savefile_mat(name, aaa, bbb);
	break;
    case FT_WAV:
	savefile_wav(name, aaa, bbb);
	break;
    case FT_NOH:
	savefile_noh(name, aaa, bbb);
	break;
    case FT_ADF:
	savefile_adf(name, aaa, bbb);
	break;
    }
    sprintf(str, " Saved file '%s'.", name);
    putmsgwk(str);
}

static void
savefile_c(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    char    str[80];
    int     nvpl = 10, nc;
    long    i, n;
    FILE   *fp;
    SIGPTR  sp;
    SIGNAL *s;

    if ((fp = wtopen(name)) == NULL)
	return;

    s = &sig[aw];
    nc = s->sgnchn;
    sp = s->sigbuf + aaa * nc;
    n = (bbb - aaa) * nc;
    rootfn(str, name);
    if (s->sgbits == BITS && s->sigzer == DACZ) {
        fprintf(fp, "static unsigned char %s[%ld] = {", str, n);
        for (i = 0; i < n; i++) {
            fprintf(fp, "%s%3d%s", (i % nvpl) ? " " : "\n    ", sp[i],
	        ((i + 1) < n) ? "," : "\n");
        }
        fprintf(fp, "};\n");
    } else {
        fprintf(fp, "short %s[%ld] = {", str, n);
        for (i = 0; i < n; i++) {
            fprintf(fp, "%s%6d%s", (i % nvpl) ? "" : "\n    ", sp[i],
	        ((i + 1) < n) ? "," : "\n");
        }
        fprintf(fp, "};\n");
    }
    fclose(fp);
}

static void
savefile_plt(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    char   *dot;
    double  rate, t, r, xgn = 0, ygn = 0, yof = 0;
    int     bits, zero, u, o, d, mn, mx, np, nc, spec, j;
    long    i, n, imn, imx, i1, i2, k, npts;
    FILE   *fp;
    SIGPTR  sp;
    SIGNAL *s;
    static char *fmt[] = {"i", "f.1", "f.2"};
    static char *ul[] = {"sec", "msec", "^m^sec", "nsec"};
    static double dt[] = {1, 1e3, 1e6, 1e9};

    if ((fp = wtopen(name)) == NULL)
	return;

    scan_signal(aw);
    n = bbb - aaa;
    s = &sig[aw];
    rate = s->adrate;
    bits = s->sgbits;
    zero = s->sigzer;
    spec = s->sigfmt;
    npts = s->siglen;
    if (rate <= 0 || n <= 0)
	return;
    t = ((double) n) / rate;
    o = limit(0, (int) (3 - log10(t)), 9);
    u = o / 3;
    d = o % 3;
    np = 500;			/* plot no more than 3*np points */
    nc = s->sgnchn;
    sp = s->sigbuf + aaa * nc;
    if ((dot = strchr(name, '.')) != NULL)
	*dot = '\0';

    fprintf(fp, "; %s - PLT file created by WAVED from '%s'\n",
	name, s->sigfnm);
    fprintf(fp, ";\n");
    if (spec) {
        xgn = (rate / 2000.0) / npts;
        ygn = 1 / ftdbsc;
        yof = ftdbof;
        r = xgn * n;
        fprintf(fp, ";    Spectra ...\n");
        fprintf(fp, ";    sample rate = %.1f Hz\n", rate);
        fprintf(fp, ";\n");
        fprintf(fp, "xlab=frequency (kHz)\n");
        fprintf(fp, "ylab=magnitude (dB)\n");
        fprintf(fp, "xper=90 : xint=%.0f.2 : xllc=1.5 : xlen=8\n", r);
        fprintf(fp, "yper=100 : yint=6.2 : yllc=1.5 : ylen=5.5\n");
        fprintf(fp, "xmin=0 : xmax=%.3f : ", r);
        fprintf(fp, "ymin=0 : ymax=120 : yhor=y : clip=y\n");
    } else {
        fprintf(fp, ";    Waveform has %ld, %d-bit, %ssigned values.\n",
            n, bits, (zero > 0) ? "un" : "");
        fprintf(fp, ";    sample rate = %.1f Hz\n", rate);
        fprintf(fp, ";    min = %d    max = %d\n", s->sigmin, s->sigmax);
        fprintf(fp, ";\n");
        fprintf(fp, "xlab=(%s)\n", ul[u]);
        fprintf(fp, "xfmt=%s : xgain=%.0f/%.0f\n", fmt[d], dt[u], rate);
        fprintf(fp, "xper=90 : xint=5.2 : xllc=1 : xlen=8.5 : xans=0\n");
        fprintf(fp, "yper=90 : yint=2.2 : yllc=3 : ylen=3   : yans=-1\n");
        fprintf(fp, "xmin=0 : xmax=%.3f\n", t * dt[u]);
        fprintf(fp, "ymin=%d : ymax=%d\n", s->sigmin, s->sigmax);
    }
    fprintf(fp, "plty=line\n");
    fprintf(fp, "tlab=%s\n", name);

    for (j = 0; j < nc; j++) {
        fprintf(fp, "data ; for channel %d\n", j + 1);
        if (n < (3 * np)) {
	    for (i = 0; i < n; i++)
                if (spec)
                    fprintf(fp, "%9.3f %9.2f\n", i * xgn, 
                        sp[i * nc + j] * ygn + yof);
                else
                    fprintf(fp, "%ld %d\n", i, sp[i * nc + j]);
        } else {
	    o = n / np;
	    d = o / 2;
	    imn = imx = 0;
	    mn = mx = sp[j];
	    for (i = 0; i < n; i++) {
	        if (mn > sp[i * nc + j])
		    mn = sp[(imn = i) * nc + j];
	        if (mx < sp[i * nc + j])
		    mx = sp[(imx = i) * nc + j];
	        if ((i % o) == 0 || i == (n - 1)) {
		    if (imn < imx) {
		        i1 = mn;
		        i2 = mx;
		    } else {
		        i1 = mx;
		        i2 = mn;
		    }
		    k = i + d;
		    if (spec) {
                        fprintf(fp, "%9.3f %9.2f\n", i * xgn, i1 * ygn + yof);
                        if (imn != imx)
                            fprintf(fp, "%9.3f %9.2f\n", k * xgn, i2 * ygn + yof);
                    } else {
                        fprintf(fp, "%ld %ld\n", i, i1);
                        if (imn != imx)
                            fprintf(fp, "%ld %ld\n", k, i2);
                    }
		    mn = mx = sp[(imn = imx = i) * nc + j];
	        }
	    }
        }
        fprintf(fp, "plot\n");
    }
    fclose(fp);
}

static void
savefile_lst(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    int     j, nc;
    long    i, ns;
    SIGPTR  sp;
    SIGNAL *s;
    FILE   *fp;

    if ((fp = wtopen(name)) == NULL)
	return;

    s = &sig[aw];
    nc = s->sgnchn;
    sp = s->sigbuf + aaa * nc;
    ns = bbb - aaa;
    fprintf(fp, "; rate=%ld\n", s->adrate);
    for (i = 0; i < ns; i++) {
        for (j = 0; j < nc; j++)
            fprintf(fp, "%7d ", sp[i * nc + j]);
        fprintf(fp, "\n");
    }
    fclose(fp);
}

static void
savefile_ils(char *name, long aaa, long bbb)
{
    int     i, length, e, fd, swap = 0, nchn;
    long    r, nn, ns;
    short  *ils_hdr;
    SIGNAL *s;
    static int format = DT_LINEAR_16;
    static long hs = 512;

    if ((fd = wbopen(name)) < 0)
	return;

    /* initialize ILS header */

    ils_hdr = (short *) buffer;
    for (i = 0; i < HDR_SIZ; i++)
	ils_hdr[i] = 0;
    s = &sig[aw];
    r = s->adrate;
    for (e = 0; r > 32000; e++)
	r /= 10;
    SF = (short) r;
    SE = e;
    nchn = s->sgnchn;
    ns = (bbb - aaa) * nchn;
    NBLKS = 1 + (ns - 1) / NSPB;
    ICHAN = 0;
    NCHAN = nchn;
    TYPE = SDF;
    INIT = 32149;

    (void) _lseek(fd, (long) 0L, 0);
    length = HDR_SIZ * 2;
    (void) _write(fd, (char *) ils_hdr, length);

    nn = s->siglen;		/* save the "real" length */
    s->siglen = ns;
    (void) _lseek(fd, (long) ILSHDSZ, 0);
    (void) _write(fd, btni_key, 6);
    (void) _write(fd, (char *) s, FNMSZ + 3 * sizeof(short));
    (void) _lseek(fd, 4L + 2 * sizeof(short), 1);
    (void) _write(fd, (char *) &(s->siglen), (4 + NMARK) * sizeof(long) + 8);
    s->siglen = nn;		/* restore the "real" length */

    write_data(fd, format, swap, hs, aaa, bbb, nchn);
}

static void
savefile_snd(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    SND_fmt *snd;
    int     bits = 16, zero = 0, fd, swap = 1, nchn;
    unsigned int len;
    SIGNAL *s;
    static int format = DT_LINEAR_16;

    if ((fd = wbopen(name)) < 0)
	return;
    s = &sig[aw];
    if (s->sgbits != bits || s->sigzer != zero)
	if (sigcon(aw, bits, zero, 0) && scrollmode == 0)
	    showdata();
    nchn = s->sgnchn;
    snd = (SND_fmt *) buffer;
    len = sizeof(SND_fmt);
    snd->ofst = len;
    snd->code = format;
    snd->size = (bbb - aaa) * 2 * nchn;
    snd->chan = nchn;
    snd->rate = s->adrate;
    snd->key = snd_key;
    sprintf(snd->info, "W%d", aw);
    exch_snd_hdr(snd);		/* change the data format */
    _write(fd, (char *) snd, len);	/* write SND header */
    write_data(fd, format, swap, (long) len, aaa, bbb, nchn);
}

static void
savefile_noh(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    int     fd, nchn = 1;
    static int format = DT_LINEAR_16;
    static long hs = 0;

    if ((fd = wbopen(name)) < 0)
	return;
    write_data(fd, format, 0, hs, aaa, bbb, nchn);
}

static void
savefile_wav(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    char    z = 0;
    int     fd, frmt, nbps, nchn;
    long    hdrsz, rifsz, fmtsz, datsz;
    WAV_fmt wfmt;

    if ((fd = wbopen(name)) < 0)
	return;
    nchn = sig[aw].sgnchn;
    wfmt.format = 1;		/* PCM */
    wfmt.channels = nchn;
    wfmt.samp_sec = sig[aw].adrate;
    if (sig[aw].sigdty == DT_UNSIGNED_8) {
        frmt = DT_UNSIGNED_8;
        nbps = 1;
    } else {
        frmt = DT_LINEAR_16;
        nbps = 2;
    }
    wfmt.block_align = nbps * nchn;
    wfmt.bits_smp = 8 * nbps;
    wfmt.byte_sec = sig[aw].adrate * nchn * nbps;
    datsz = (bbb - aaa) * nchn * nbps;
    fmtsz = sizeof(wfmt);
    hdrsz = 7 * 4 + fmtsz;
    rifsz = 5 * 4 + fmtsz + datsz + (datsz % 2);
    _write(fd, "RIFF", 4);
    _write(fd, (char *) &rifsz, 4);
    _write(fd, "WAVE", 4);
    _write(fd, "fmt ", 4);
    _write(fd, (char *) &fmtsz, 4);
    _write(fd, (char *) &wfmt, (unsigned) fmtsz);
    _write(fd, "data", 4);
    _write(fd, (char *) &datsz, 4);
    write_data(fd, frmt, 0, hdrsz, aaa, bbb, nchn);
    if (datsz % 2)
        _write(fd, (char *) &z, 1);
}

static int
wr_mat_hdr(int fd, int m, int p, char *nam, long ns, int nc)
{
    int     len;
    long    hdr[5];

    len = strlen(nam) + 1;
/*
 *  matlab header -
 *	type = MOPT where:
 *	M=0 for pc; 1 for sun
 *	O=0 col or 1 row;
 *	P= [0,1,2,3,4]=>[r*8,r*4,I*4,I*2,I*4 unsigned integers]
 *	T= 0 for matrix or 1 for text (stored as r*4 numbers 0<i<255).
 */
    hdr[0] = m * 1000 + p * 10; /* MOPT */
    hdr[1] = nc;		/* rows */
    hdr[2] = ns;		/* cols */
    hdr[3] = 0;			/* 0 for real, 1 for complex */
    hdr[4] = (long) len;	/* name length (including null byte) */
    (void) _write(fd, (char *) hdr, 20);
    (void) _write(fd, (char *) nam, len);

    return (20 + len);
}

static void
savefile_mat(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    char    nam[FNSZ];
    double  rate;
    int     fd, nchn;
    long    ns, hs = 0;
    static int format = DT_LINEAR_16;

    if ((fd = wbopen(name)) < 0)
	return;
    rate = sig[aw].adrate;
    nchn = sig[aw].sgnchn;
    hs += wr_mat_hdr(fd, cpu_lsb, 0, "rate", 1L, 1) + sizeof(double);
    (void) _write(fd, (char *) (&rate), sizeof(double));
    ns = bbb - aaa;
    sprintf(nam, "W%d", aw);
    hs += wr_mat_hdr(fd, cpu_lsb, 3, nam, ns, nchn);
    write_data(fd, format, 0, hs, aaa, bbb, nchn);
}

static void
savefile_adf(name, aaa, bbb)
char   *name;
long    aaa, bbb;
{
    ADF_fmt *adf_hdr;
    int     fd, swap = 0, nchn;
    SIGNAL *s;
    static int format = DT_LINEAR_16;
    static long hs = sizeof(ADF_fmt);

    if ((fd = wbopen(name)) < 0)
	return;

    s = &sig[aw];
    nchn = s->sgnchn;
    adf_hdr = (ADF_fmt *) buffer;
    strncpy((char *) adf_hdr, adf_key, 8);
    adf_hdr->npts = bbb - aaa;
    adf_hdr->chan = nchn;
    adf_hdr->bits = 16;
    adf_hdr->code = 0;
    adf_hdr->rate = (float) (s->adrate * 0.001);
    adf_hdr->peak = 0;
    adf_hdr->gain1 = 0;
    adf_hdr->gain2 = 0;
    adf_hdr->vrms = 0;
    adf_hdr->dbref = 0;

    (void) _write(fd, (char *) adf_hdr, (int) hs);
    write_data(fd, format, swap, hs, aaa, bbb, nchn);
}

static FILE   *
wtopen(name)
char   *name;
{
    char    str[80];
    FILE   *fp;

    fp = fopen(name, "w");
    if (fp == NULL) {
	sprintf(str, " can't open file '%s' ", name);
	putmsgwk(str);
    } else {
	sprintf(str, " writing (%s) output to file '%s' ", ft[file_type], name);
	putmsg(str);
    }
    return (fp);
}

static int
wbopen(name)
char   *name;
{
    char    str[80];
    int     fd;

    fd = _open(name, O_MODE, P_MODE);
    if (fd < 0) {
	sprintf(str, " can't open file '%s' ", name);
	putmsgwk(str);
    } else {
	sprintf(str, " writing (%s) output to file '%s' ", ft[file_type], name);
	putmsg(str);
    }
    return (fd);
}

static void
write_data(fd, format, swap, offset, aaa, bbb, nchn)
long    offset, aaa, bbb;
{
    unsigned char *bp;
    int     j, np, nb, nw, nc;
    long    i, ns;
    SIGPTR  sp;
    SIGNAL *s;

    s = &sig[aw];
    nc = limit(1, nchn, MAX_CHAN);
    ns = (bbb - aaa) * nc;
    bp = (unsigned char *) buffer;
    nw = sizeof(buffer) / sizeof(short);
    sp = s->sigbuf + aaa * nc;
    (void) _lseek(fd, offset, 0);
    for (i = 0; i < ns; i += nw) {
	np = ((i + nw) > ns) ? (ns - i) : nw;
	switch (format) {
	case DT_UNSIGNED_8:
	    for (j = 0; j < np; j++)
		bp[j] = (unsigned char) sp[i + j];
	    (void) _write(fd, (char *) buffer, np);
	    break;
	case DT_LINEAR_16:
	    for (j = 0; j < np; j++)
		buffer[j] = sp[i + j];
	    nb = 2 * np;
	    if (swap)
		_swab((char *) buffer, (char *) buffer, nb);
	    (void) _write(fd, (char *) buffer, nb);
	    break;
	}
    }
    _close(fd);
}

static void
remext(fn)
char   *fn;
{
    char   *dot;

    if ((dot = strrchr(fn, '.')) != NULL)
	*dot = '\0';
}

static void
defext(fn)
char   *fn;
{
    static char *ex[NFT] = {
    	".ils", ".c", ".trn", ".lst", ".plt", ".snd", ".", ".mat", 
    	".wav", ".tok", ".adf"
    };

    if (strchr(fn, '.') == NULL)
	strcat(fn, ex[file_type]);
}
