/* window.c */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "waved.h"
#include "video.h"

#define ESC 27
#define TOP 0
#define BOT (msgy - ln_sp)
#define SCRPOS(x)   (scrp0 + (((x) * scrp1) / scrp2))

void    outline_box();
void    title_box();
void    signal_box();
void    place_cursor();
void    showscrollbox();
void    activate_window();
void    curstime();
void    g_cursor_del();
void    g_cursor_init();
void    g_cursor_move();
void    g_sb_del();
void    g_sb_move();
void    g_drawcmag();
void    g_drawwave();
void    windinit();
void    goto_samp();

extern int ln_sp, ch_wd, ch_ht;

static char *ct_tu;
static int ct_dp;
static float ct_dt;
static int curs_y, curs_h;
static long scrp0 = 0, scrp1 = 0, scrp2 = 1;
static int markpix[NMARK];

void
init_wind()
{
    int     i;

    clpwin = 0;
    fsw = 1;
    lsw = NWIND - 1;
    nsw = 1 + lsw - fsw;
    aw = fsw;
    for (i = 0; i < NWIND; i++)
	windinit(i);
}

void
windinit(int num)
{
    int     j;

    win[num].winx1 = 0;
    win[num].winx2 = scrszx - 1;
    win[num].winmrk1 = NOMARK;
    win[num].winmrk2 = NOMARK;

    if (num == clpwin)
	strcpy(sig[clpwin].sigfnm, "[clip]");
    else
	sig[num].sigfnm[0] = '\0';
    sig[num].sigtom = 0;
    sig[num].sigmin = 0;
    sig[num].sigmax = 0;
    sig[num].sigbuf = (SIGPTR) NULL;
    sig[num].sigzer = 0;
    sig[num].sgbits = 0;
    sig[num].siglen = 0;
    sig[num].sgtime = 0;
    sig[num].adrate = def_rate;
    sig[num].darate = def_rate;
    for (j = 0; j < 26; j++)
	sig[num].sigmrk[j] = NOMARK;
    sig[num].sigmrk[HOME_MARK] = 0;
    sig[num].sigmrk[END_MARK] = 0;
    sig[num].sigmrk[CURSOR_MARK] = 0;
}

void
windpos(wn)
{
    int     h, y, tw;
    long    ns;
    WINDOW *w;

    y = BOT - TOP;
    nsw = 1 + lsw - fsw;
    h = limit(ln_sp, y / nsw, y / 2);
    w = &win[wn];
    if (maximize) {
	w->winy1 = TOP;
	w->winy2 = w->winy1 + y - 1;
    } else {
	w->winy1 = TOP + (wn - fsw) * h;
	w->winy2 = w->winy1 + h - 1;
    }

    ns = sig[wn].siglen;
    if (w->winmrk1 == NOMARK)
	w->winmrk1 = 0;
    if (w->winmrk2 == NOMARK)
	w->winmrk2 = ns;
    w->winmrk1 = limit(0, w->winmrk1, ns);
    w->winmrk2 = limit(0, w->winmrk2, ns);
    if (w->winmrk2 < w->winmrk1) {
	tw = w->winmrk2;
	w->winmrk1 = w->winmrk2;
	w->winmrk2 = tw;
    }
}

void
showwindow(int wn)
{
    windpos(wn);
    if (win[wn].winy1 >= TOP && win[wn].winy2 < BOT) {
	if (wn == aw) {
	    activate_window(wn);
	    outline_box(wn, active);
	    title_box(wn);
	    signal_box(wn);
	    place_cursor(wn);
	} else if (!maximize) {
	    outline_box(wn, boxcol);
	    title_box(wn);
	    signal_box(wn);
	}
    }
    (void) chkevent(); /* updates windows */
}

void
outline_box(wn, color)
{
    int     xx1, xx2, yy1, yy2;
    WINDOW *w;

    w = &win[wn];

    xx1 = w->winx1 + 1;
    xx2 = w->winx2 - 1;
    yy1 = w->winy1 + 1;
    yy2 = w->winy1 + 1 + ln_sp;
    g_box(xx1, yy1, xx2, yy2, color);
    yy1 = yy2 + 1;
    yy2 = w->winy2 - 1;
    g_box(xx1, yy1, xx2, yy2, color);
    g_box(w->winx1, w->winy1, w->winx2, w->winy2, color);

    if ((wn == lsw) && (w->winy2 < (BOT - 1))) {
	xx1 = w->winx1;
	yy1 = w->winy2 + 1;
	xx2 = w->winx2;
	yy2 = BOT - 1;
	g_bar(xx1, yy1, xx2, yy2, scbgc);
    }
}

void
title_box(wn)
{
    char    label[80];
    char   *un, *msg;
    int     dp;
    int     xx1, xx2, yy1, yy2;
    float   dt, wd, df;
    WINDOW *w;
    SIGNAL *s;

    w = &win[wn];
    s = &sig[wn];
    xx1 = w->winx1 + 1;
    yy1 = w->winy1 + 1;
    xx2 = w->winx2 - 1;
    yy2 = w->winy1 + 1 + ln_sp;
    g_bar(xx1 + 1, yy1 + 1, xx2 - 1, yy2 - 1, blbgc);

    sprintf(label, "%c %d : %.*s ", (wn == aw) ? '>' : ' ', wn, FNMSZ, s->sigfnm);
    if (label[0] != '\0') {
	g_settextclr(blfgc, blbgc);
	g_text(w->winx1 + 2, w->winy1 + 2 + ch_ht, label);
    }
    msg = label;
    if (s->siglen <= 0) {
	sprintf(msg, "No signal");
    } else {
	s->sigtom = (int) log10((s->siglen * 1.0e6) / s->adrate);
	if (s->sigtom < 6) {
	    dp = 7 - s->sigtom;
	    dt = (float) (1000.0 / s->adrate);
	    un = "ms";
	} else {
	    dp = 10 - s->sigtom;
	    dt = (float) (1.0 / s->adrate);
	    un = "s ";
	}
        if (s->sigfmt == 1) {
            df = (float) ((0.001 * sig[aw].adrate) / (2 * s->siglen));
            wd = (w->winmrk2 - w->winmrk1) * df;
            sprintf(msg, "Bandwidth = %6.*f %s  ", 3, wd, "kHz");
        } else {
            wd = (w->winmrk2 - w->winmrk1) * dt;
            sprintf(msg, "Duration = %6.*f %s  ", dp, wd, un);
        }
    }

    g_settextclr(bmfgc, bmbgc);
    dp = strlen(msg);		/* avoid MSC compiler error */
    cumx = w->winx2 - (dp + 1) * ch_wd;
    cumy = w->winy1 + 2 + ch_ht;
    g_text(cumx, cumy, msg);
}

void
signal_box(wn)
{
    int     xpix, ypix, xofst, yofst;
    int     xx1, xx2, yy1, yy2, nch, i, dy;
    long    npt, rng, zer, smn, smx;
    WINDOW *w;
    SIGPTR  pbf;
    SIGNAL *s;

    w = &win[wn];
    s = &sig[wn];

    xx1 = w->winx1 + 1;
    yy1 = w->winy1 + 1 + ln_sp;
    xx2 = w->winx2 - 1;
    yy2 = w->winy2 - 1;
    g_bar(xx1 + 1, yy1 + 1, xx2 - 1, yy2 - 1, wnbgc);

    if (s->siglen <= 0)
	return;
    s->sigmrk[HOME_MARK] = 0;
    s->sigmrk[END_MARK] = s->siglen;

    smn = w->winmrk1;
    smx = w->winmrk2;
    npt = smx - smn;
    nch = limit(1, s->sgnchn, MAX_CHAN);
    if (npt > 0) {
        pbf = s->sigbuf + smn * nch;
        rng = 1L << (s->sgbits);
        zer = s->sigzer;
	xpix = xx2 - xx1 - 4;	    /* allow for markers at the endpoints */
	if (w->winmod & 1) {
            ypix = (yy2 - yy1 - 4) / nch;
            dy = ypix;
        } else {
            ypix = yy2 - yy1 - 4;
            dy = 0;
        }
	xofst = xx1 + 2;
        yofst = yy1 + 2 + ypix / 2;
        for (i = 0; i < nch; i++)
            g_drawwave(pbf + i, npt, nch, xpix, ypix, xofst, 
                yofst + i * dy, rng, zer, wavcol - i);
    }
}

void
place_cursor(wn)
{
    int     c, col, top;
    WINDOW *w;
    SIGNAL *s;

    w = &win[wn];
    s = &sig[wn];
    if (s->siglen <= 0) {
	clrsts();
	return;
    }

    /* Define the screen position parameters */

    scrp0 = w->winx1 + 3;
    scrp1 = w->winx2 - w->winx1 - 6;
    scrp2 = (cursmax > cursmin) ? cursmax - cursmin : 1;

    /* draw the box indicating the location of the scroll window */

    showscrollbox(wn);

    /* Now put out any markers that have been defined */

    top = w->winy2 - ln_sp - 4;
    for (c = 0; c < 26; c++) {
	if (s->sigmrk[c] < 0 || s->sigmrk[c] > s->siglen)
	    s->sigmrk[c] = NOMARK;
	if (s->sigmrk[c] != NOMARK) {
	    col = SCRPOS(s->sigmrk[c] - cursmin);
	    g_drawmark(col, top, c);
	    markpix[c] = col;
	}
    }

    /* draw cursor */

    w->wincp = curspnt;
    col = SCRPOS(curspnt - cursmin);
    g_cursor_init(curs_h, csrcol);
    g_cursor_move(col, curs_y);

    curstime();
}

void
showscrollbox(wn)
{
    int     xx1, xx2, yy1, yy2;
    long    scrolpnt;
    WINDOW *w;
    SIGNAL *s;

    if (scrollbox == 0) {
	g_sb_del();
	return;
    }
    w = &win[wn];
    s = &sig[wn];
    if (s->siglen < scrszx)
	return;
    scrolpnt = limit(0, curspnt - scroff, s->siglen - scrszx);
    xx1 = SCRPOS(scrolpnt - cursmin) - 1;
    xx2 = SCRPOS(scrolpnt + scrszx - 1 - cursmin) + 1;
    yy1 = w->winy1 + 1 + ln_sp;
    yy2 = w->winy2 - 2;
    if (xx2 > w->winx1 && xx1 < w->winx2) {
	if (xx1 < w->winx1)
	    xx1 = w->winx1;
	if (xx2 > w->winx2)
	    xx2 = w->winx2;
	g_sb_move(xx1, yy1, xx2, yy2, scwcol);
    }
}

void
activate_window(wn)
{
    WINDOW *w;
    SIGNAL *s;

    w = &win[wn];
    s = &sig[wn];
    if (s->siglen <= 0)
	return;
    s->sigmrk[HOME_MARK] = 0;
    s->sigmrk[END_MARK] = s->siglen;
    cursmin = w->winmrk1;
    cursmax = w->winmrk2;
    if (curspnt < cursmin || curspnt > cursmax)
	curspnt = (cursmin + cursmax) / 2;

    /* Verify signal definintion */

    if (s->adrate <= 0)
	s->adrate = def_rate;
    s->sigtom = (int) log10((s->siglen * 1.0e6) / s->adrate);

    if (s->siglen >= scrszx) {
	scroff = scrszx / 2;
    } else {
	scroff = s->siglen / 2;
	scrollbox = 0;
    }
    g_sb_del();

    /* Initialize the cursor */

    g_cursor_del();
    curs_y = w->winy1 + ln_sp + 2;
    curs_h = w->winy2 - curs_y - 2;
    if (s->sigtom < 6) {
	ct_dp = 7 - s->sigtom;
	ct_dt = (float) (1000.0 / s->adrate);
	ct_tu = "ms";
    } else {
	ct_dp = 10 - s->sigtom;
	ct_dt = (float) (1.0 / s->adrate);
	ct_tu = "s ";
    }
}

/* CURSTIME to put up the time at the cursor position */

void
curstime()
{
    char    sts[80];
    double  ct, td, df, db, db1, db2;
    int     cc, nc;
    long    ns;

    ns = sig[aw].siglen;
    nc = sig[aw].sgnchn;
    if (sig[aw].sigfmt == 1) {
        df = (0.001 * sig[aw].adrate) / (2 * ns);
        sprintf(sts, " FFT Cursor ");
        cc = strlen(sts);
        sprintf(sts + cc, "at %6.3f kHz, ", df * curspnt);
        cc = strlen(sts);
        if (nc < 2) {
            db = (sig[aw].sigbuf[curspnt * nc] / ftdbsc) + ftdbof;
            sprintf(sts + cc, "Value = %5.1f dB  ", db);
        } else {
            db1 = (sig[aw].sigbuf[curspnt * nc] / ftdbsc) + ftdbof;
            db2 = (sig[aw].sigbuf[curspnt * nc + 1] / ftdbsc) + ftdbof;
            sprintf(sts + cc, "Value = %5.1f,%5.1f dB  ", db1, db2);
        }
        putsts(sts);
    } else {
        ct = ct_dt * curspnt;
        td = ct_dt * ns;
        sprintf(sts, " Cursor at  ");
        cc = strlen(sts);
        sprintf(sts + cc, "%6.*f/%6.*f %s, ", ct_dp, ct, ct_dp, td, ct_tu);
        cc = strlen(sts);
        sprintf(sts + cc, "(%6ld/%6ld).  ", curspnt, ns);
        cc = strlen(sts);
        sprintf(sts + cc, "Value = %6d  ", sig[aw].sigbuf[curspnt * nc]);
        putsts(sts);
    }
}

void
update_cursor()
{
    int     curscol;

    if (sig[aw].siglen <= 0) {
        g_cursor_del();
        g_sb_del();
	clrsts();
    } else {
        curscol = SCRPOS(curspnt - cursmin);
        g_cursor_move(curscol, curs_y);
        curstime();
    }
}

void
update_mark(int c, long p)
{
    int     col, top;

    top = win[aw].winy2 - 20;
    if (sig[aw].sigmrk[c] != NOMARK) {
	col = markpix[c];
	g_drawmark(col, top, c);
    }
    sig[aw].sigmrk[c] = p;
    if (sig[aw].sigmrk[c] != NOMARK) {
	col = SCRPOS(p - cursmin);
	g_drawmark(col, top, c);
	markpix[c] = col;
    }
}

void
chaw(int wn)
{
    int     c, col, top, ow;

    if (wn < 0 || wn >= NWIND || wn == aw)
	return;

    ow = aw;
    g_sb_del();
    g_cursor_del();
    top = win[ow].winy2 - ln_sp - 4;
    for (c = 0; c < NMARK; c++)
	if (sig[ow].sigmrk[c] != NOMARK) {
	    col = markpix[c];
	    g_drawmark(col, top, c);
	}
    win[ow].wincp = curspnt;

    aw = wn;
    curspnt = win[aw].wincp;
    if (!maximize) {
	outline_box(ow, boxcol);
    	if (sig[ow].siglen > 0)
	    title_box(ow);
    }
    windpos(aw);
    activate_window(aw);
    outline_box(aw, active);
    title_box(aw);
    if (maximize || sig[aw].siglen == 0)
	signal_box(aw);
    place_cursor(aw);
}

void
goto_samp(long cp)
{
    long    npts, nsmp, mrk1, mrk2;

    nsmp = sig[aw].siglen;
    curspnt = limit(0, cp, nsmp);
    mrk1 = win[aw].winmrk1;
    mrk2 = win[aw].winmrk2;
    npts = mrk2 - mrk1;
    if (cp < mrk1) {
	mrk1 = limit(0, cp - npts / 2, nsmp);
	mrk2 = limit(0, mrk1 + npts, nsmp);
    }
    if (cp > mrk2) {
	mrk2 = limit(0, cp + npts / 2, nsmp);
	mrk1 = limit(0, mrk2 - npts, nsmp);
    }
    if (mrk1 == win[aw].winmrk1 && mrk2 == win[aw].winmrk2) {
	update_cursor();
    } else {
	win[aw].winmrk1 = mrk1;
	win[aw].winmrk2 = mrk2;
	showdata();
    }
    cursmin = mrk1;
    cursmax = mrk2;
}
